#!/usr/bin/env python
"""Script to run a modeller that sends tasks off to Pods."""
import logging
from os import PathLike
from pathlib import Path
from typing import Union

import fire

from bitfount import config
from bitfount.runners.modeller_runner import (
    DEFAULT_MODEL_OUT,
    run_modeller,
    setup_modeller_from_config_file,
)
from bitfount.runners.utils import setup_loggers

loggers = setup_loggers([logging.getLogger("bitfount")])

config._BITFOUNT_CLI_MODE = True


def run(
    path_to_config_yaml: Union[str, PathLike],
    require_all_pods: bool = False,
    model_out: Path = DEFAULT_MODEL_OUT,
) -> None:
    """Runs a modeller from a config file."""
    modeller, pod_identifiers = setup_modeller_from_config_file(path_to_config_yaml)
    run_modeller(modeller, pod_identifiers, require_all_pods, model_out)


def main() -> None:
    """Script entry point."""
    fire.Fire(run)


if __name__ == "__main__":
    main()



"""Base mailbox classes for other mailbox classes to inherit from."""
import asyncio
from asyncio import FIRST_COMPLETED, AbstractEventLoop, Task
from asyncio.futures import Future as AsyncFuture
from collections import defaultdict
from concurrent.futures import Future as ConcurrentFuture
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
import inspect
import threading
import time
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    DefaultDict,
    Dict,
    Final,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)
import warnings

import async_timeout
from grpc import RpcError, StatusCode
import msgpack

from bitfount.config import _BITFOUNT_MULTITHREADING_DEBUG
from bitfount.exceptions import BitfountError
from bitfount.federated.exceptions import (
    MessageHandlerNotFoundError,
    MessageRetrievalError,
    MessageTypeSpecificHandlerNotFoundError,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.transport.handlers import _PriorityHandler, _TemporaryHandler
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _MessageEncryption,
    _MessageService,
    msgpackext_decode,
    msgpackext_encode,
)
from bitfount.utils.concurrency_utils import (
    NO_POLLING,
    ThreadWithException,
    _synchronised,
    async_lock,
    asyncnullcontext,
    await_threading_event,
)

logger = _get_federated_logger(__name__)

SyncHandler = Callable[[_BitfountMessage], None]
AsyncHandler = Callable[[_BitfountMessage], Awaitable[None]]
NonPriorityHandler = Union[SyncHandler, AsyncHandler]
Handler = Union[NonPriorityHandler, _PriorityHandler]

# The window in which handlers can be registered for them to be dispatched for a
# given message.
HANDLER_RETRIEVAL_TIMEOUT: int = 30

# Additional types for handler registration storage
ANY_MESSAGE: Final = "ANY_MESSAGE"
_ExtendedMessageTypes = Union[_BitfountMessageType, Literal["ANY_MESSAGE"]]
# Registered handlers can additionally be of type _TemporaryHandler
_RegisterHandler = Union[Handler, _TemporaryHandler]
_HandlersDict = Dict[_RegisterHandler, Optional[asyncio.Lock]]


class _HandlerLockPair(NamedTuple):
    """Pair of handler and associated exclusivity lock."""

    handler: _RegisterHandler
    lock: Optional[asyncio.Lock]


class _HandlerRegistry:
    """Registry for message handlers.

    Thread-safe.
    """

    _registry: DefaultDict[_ExtendedMessageTypes, _HandlersDict]
    _sync_lock: threading.RLock

    def __init__(
        self,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ) -> None:
        self._registry: DefaultDict[_ExtendedMessageTypes, _HandlersDict] = defaultdict(
            dict
        )
        if handlers:
            # Register supplied handlers
            for message_type, handlers_ in handlers.items():
                try:
                    # Assume iterable
                    for handler in cast(Iterable[Handler], handlers_):
                        # TODO: [BIT-1048] Revisit this and decide if _all_ messages
                        #       need to be exclusive in this way.
                        self.register_handler(message_type, handler, exclusive=True)
                except TypeError:
                    # Otherwise, register individual handler
                    self.register_handler(
                        message_type, cast(Handler, handlers_), exclusive=True
                    )

        # Lock for managing multithread operations
        self._sync_lock = threading.RLock()

    @_synchronised
    def register_handler(
        self,
        message_type: _ExtendedMessageTypes,
        handler: _RegisterHandler,
        exclusive: bool = True,
    ) -> None:
        """Registers a handler for a specific message type.

        Args:
            message_type: The message type to register the handler for.
            handler: The handler.
            exclusive: Whether only a single instance of the handler can be running
                at a given time.
        """
        if exclusive:
            if isinstance(handler, _PriorityHandler):
                # The exclusivity locking is handled within the priority handler
                # instead
                lock = None
                handler.set_exclusive()
            else:
                lock = asyncio.Lock()
        else:
            lock = None

        self._registry[message_type][handler] = lock

    @_synchronised
    def delete_handler(
        self,
        message_type: _ExtendedMessageTypes,
        handler: _RegisterHandler,
    ) -> None:
        """Deletes a handler associated with the message type."""
        self._registry[message_type].pop(handler, None)  # avoids KeyError

    @_synchronised
    def delete_all_handlers(self, message_type: _ExtendedMessageTypes) -> None:
        """Deletes all handlers for a specific message type."""
        self._registry[message_type].clear()

    @overload
    def get_handlers(
        self, message_type: _BitfountMessageType, with_locks: Literal[True]
    ) -> List[_HandlerLockPair]:
        """Gets all handlers for a specific message type."""
        ...

    @overload
    def get_handlers(
        self, message_type: _ExtendedMessageTypes, with_locks: Literal[False] = ...
    ) -> List[_RegisterHandler]:
        """Gets all handlers for a specific message type."""
        ...

    @_synchronised  # type: ignore[misc] # Reason: https://github.com/python/mypy/issues/12716 # noqa: B950
    def get_handlers(
        self, message_type: _ExtendedMessageTypes, with_locks: bool = False
    ) -> Union[List[_RegisterHandler], List[_HandlerLockPair]]:
        """Gets all handlers for a specific message type.

        If `with_locks` is True, returns a list of tuples of the handlers and their
        exclusivity locks if present.

        :::caution

        This does not retrieve the universal handlers. If you want to retrieve those
        as well you must call get_handlers(ANY_MESSAGE) and merge the two handler
        lists together.

        :::

        Args:
            message_type: The message type to retrieve the handlers for.
            with_locks: Whether to include handlers' exclusivity locks.

        Returns:
            Either a list of handlers or a list of handler-lock tuples.
        """
        handlers: Union[List[_RegisterHandler], List[_HandlerLockPair]]
        if not with_locks:
            handlers = list(self._registry[message_type].keys())
        else:
            handlers = [
                _HandlerLockPair(k, v) for k, v in self._registry[message_type].items()
            ]

        return handlers

    @_synchronised
    def get_lock(
        self,
        message_type: _ExtendedMessageTypes,
        handler: _RegisterHandler,
    ) -> Optional[asyncio.Lock]:
        """Returns the exclusivity lock for a given handler and message type.

        Returns:
            The exclusivity lock for the handler or None if no lock.

        Raises:
            KeyError: If handler is not associated with message type.
        """
        return self._registry[message_type][handler]


class _BaseMailbox:
    """The base mailbox class.

    Contains handlers and message service.

    Args:
        mailbox_id: the ID of the mailbox to monitor.
        message_service: underlying message service instance.
        handlers: an optional mapping of message types to handlers to initialise with.
    """

    def __init__(
        self,
        mailbox_id: str,
        message_service: _MessageService,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ):
        self.mailbox_id = mailbox_id
        self.message_service = message_service

        self._handlers: _HandlerRegistry = _HandlerRegistry(handlers)

        # Only one call to _listen_for_messages() should be allowed at a time.
        # Otherwise we run the risk of messages being pulled off of the mailbox
        # by a listener that doesn't have the right handlers. Each mailbox should
        # only need one listener as it runs indefinitely.
        self._listening_lock = threading.Lock()

        # To enable a smart back-off when no handler is found before reattempting
        # we introduce an event that can monitor when new handlers are added. This
        # allows us to handle the situation where a response comes through faster
        # than the correct handler can be attached.
        self._new_handler_added_monitors_lock = threading.Lock()
        self._new_handler_added_monitors: Set[threading.Event] = set()

    async def log(self, message: Mapping[str, object]) -> None:
        """Log message to remote task participant."""
        raise NotImplementedError

    def _setup_federated_logging(self) -> None:
        """Sets up federated logging."""
        raise NotImplementedError

    @contextmanager
    def listen(
        self, handler_dispatch_loop: Optional[AbstractEventLoop] = None
    ) -> Generator[threading.Thread, None, None]:
        """Starts the mailbox listening for messages in a separate thread.

        High-priority messages will be dispatched to a thread pool to be executed
        immediately. Low-priority messages will be dispatched to the event loop
        supplied to this method.

        Args:
            handler_dispatch_loop: The event loop to dispatch low-priority message
                handling to. If not supplied the running event loop from the calling
                thread is used.

        Yields:
            The thread that the mailbox listener is running in.

        Raises:
            BitfountError: If no handler_dispatch_loop is supplied and no event
                loop is running.
        """
        if not handler_dispatch_loop:
            try:
                handler_dispatch_loop = asyncio.get_running_loop()
            except RuntimeError as ex:
                raise BitfountError(
                    "Attempted to run mailbox listener without a running event loop."
                ) from ex

        # Create stop event, so we can notify the underlying thread to cease
        stop_event = threading.Event()

        mailbox_thread = ThreadWithException(
            target=self._start_in_own_thread,
            name=f"mailbox_listener_thread_{self.mailbox_id}",
            args=(handler_dispatch_loop, stop_event),
            daemon=True,
        )

        try:
            mailbox_thread.start()
            yield mailbox_thread
        finally:
            logger.debug(f"Stopping mailbox listener ({self.mailbox_id})...")
            stop_event.set()

            if mailbox_thread.is_alive():
                stop_timeout = 15
                logger.debug(
                    f"Waiting up to {stop_timeout} seconds"
                    f" for mailbox listener ({self.mailbox_id}) to stop..."
                )

                mailbox_thread.join(timeout=stop_timeout)

                if mailbox_thread.is_alive():
                    raise TimeoutError(
                        f"Mailbox listener ({self.mailbox_id}) did not stop in time."
                    )

            logger.debug(f"Mailbox listener ({self.mailbox_id}) stopped.")

    def _start_in_own_thread(
        self, handler_dispatch_loop: AbstractEventLoop, stop_event: threading.Event
    ) -> None:
        """Start listening for messages in new event loop.

        This is designed to be a thread target.
        """
        return asyncio.run(self._listen_for_messages(handler_dispatch_loop, stop_event))

    async def listen_indefinitely(
        self, handler_dispatch_loop: Optional[AbstractEventLoop] = None
    ) -> None:
        """Listen to a mailbox indefinitely without blocking the event loop.

        Does this by setting up a mailbox listener and dispatcher in a separate
        thread and then providing an async Future to await on. If the listener finishes
        or the caller of this function is cancelled then this will gracefully exit out,
        otherwise it allows the listener to run indefinitely.

        Args:
            handler_dispatch_loop: The event loop to dispatch low-priority message
                handling to. If not supplied the running event loop from the calling
                thread is used.

        Raises:
            BitfountError: If no handler_dispatch_loop is supplied and no event
                loop is running.
        """
        with self.listen(handler_dispatch_loop) as listen_thread:
            # We can't use the ThreadPoolExecutor context manager here because
            # of how the target thread functions and our intentions.
            #
            # If the process is cancelled then we want this function to exit out
            # which will cause the listen_thread to be gracefully shutdown as we
            # leave the self.listen() context manager. If we use the ThreadPoolExecutor
            # context manager it calls executor.shutdown(wait=True) which will cause
            # a deadlock as listen_thread.join() (and hence ThreadPoolExecutor)
            # won't finish until we're out of the self.listen() context manager,
            # but the self.listen() context manager won't finish until we leave
            # the ThreadPoolExecutor context manager.
            #
            # Instead, we need to manually call shutdown on the executor so that
            # it won't block, knowing that the thread in the executor will finish
            # soon because of the graceful shutdown mechanism in self.listen().
            executor = ThreadPoolExecutor(
                1, thread_name_prefix="listen_indefinitely_executor"
            )
            try:
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(executor, listen_thread.join)
            finally:
                executor.shutdown(wait=False)
        return None

    async def _listen_for_messages(
        self,
        handler_dispatch_loop: AbstractEventLoop,
        stop_event: threading.Event,
    ) -> None:
        """Listens for messages on the target mailbox.

        Received messages are passed to the relevant handlers. If no relevant
        handlers are found, it will wait for up to `HANDLER_RETRIEVAL_TIMEOUT`
        for one to be registered. This avoids the situation of a response coming
        through faster than a handler can be registered. If not, it is passed
        to the default handler.
        """
        running_dispatch_tasks: Set[Task] = set()
        done_dispatch_tasks: Set[Task] = set()

        # Guarantee that only one listener is listening at a time.
        # We can set there to be NO_POLLING despite the dangers as
        # `_listen_for_messages()` should only be called once for each mailbox.
        async with async_lock(
            self._listening_lock,
            lock_name=f"listening_lock_{self.mailbox_id}",
            polling_timeout=NO_POLLING,
        ):
            try:
                async for message in self.message_service.poll_for_messages(
                    self.mailbox_id,
                    stop_event,
                ):
                    if stop_event.is_set():
                        logger.info(
                            f"Stopping listening for messages"
                            f" in mailbox {self.mailbox_id}"
                        )
                        break

                    # Create dispatch task to run in background and handle
                    # maintaining a reference to it. Also add callbacks to
                    # move it between sets once it is done.
                    dispatch_task = asyncio.create_task(
                        self._dispatch_message_handlers(message, handler_dispatch_loop),
                        name=f"{message.message_type.name}_handler_dispatch_{self.mailbox_id}",
                    )
                    running_dispatch_tasks.add(dispatch_task)
                    dispatch_task.add_done_callback(running_dispatch_tasks.discard)
                    dispatch_task.add_done_callback(done_dispatch_tasks.add)

                    # Check on any "done" handler dispatch tasks
                    # TODO: [NO_TICKET: Idle thoughts rather than actionable]
                    #       What to do with the awaitables from the dispatch; do
                    #       they need weakref protection?
                    #       See: https://github.com/python/cpython/issues/88831
                    #       See: https://github.com/python/asyncio/issues/397#issuecomment-339014739  # noqa: B950
                    examined_tasks = set()
                    for task in done_dispatch_tasks:
                        # If there was an exception in the task dispatch,
                        # this will reraise it here.
                        task.result()
                        logger.debug(
                            f"Dispatch task {task.get_name()} is done"
                            f" in mailbox {self.mailbox_id}"
                        )
                        examined_tasks.add(task)
                    # Remove examined tasks from the done set
                    done_dispatch_tasks -= examined_tasks

            # General message service issues to log out before failing.
            except RpcError as err:
                if err.code() == StatusCode.UNAVAILABLE:
                    logger.warning("Message Service unavailable")
                if err.code() == StatusCode.UNAUTHENTICATED:
                    # This could be a temporary token expiry issue
                    logger.info(
                        f"Authentication to read from '{self.mailbox_id}' failed"
                    )
                if err.code() == StatusCode.PERMISSION_DENIED:
                    logger.warning(
                        f"You don't own a pod with the mailbox: {self.mailbox_id}. "
                        f"Ensure it exists on Bitfount Hub."
                    )
                if err.code() == StatusCode.FAILED_PRECONDITION:
                    logger.debug(
                        f"No mailbox exists for '{self.mailbox_id}', "
                        f"ensure connect_pod() or send_task_requests() is called first."
                    )
                raise MessageRetrievalError(
                    f"An error occurred when trying to communicate"
                    f" with the messaging service: {err}"
                )
            finally:
                # Cancel outstanding dispatch tasks
                if len(running_dispatch_tasks) > 0:
                    cancelled_tasks: Set[str] = set()

                    for task in running_dispatch_tasks:
                        task.cancel()
                        cancelled_tasks.add(task.get_name())

                    logger.warning(
                        f"Message listener ended with running handler dispatches."
                        f" Cancelled: {', '.join(sorted(cancelled_tasks))}"
                    )

    async def _dispatch_message_handlers(
        self,
        message: _BitfountMessage,
        handler_dispatch_loop: AbstractEventLoop,
        timeout: int = HANDLER_RETRIEVAL_TIMEOUT,
    ) -> List[Union[ConcurrentFuture, _PriorityHandler]]:
        """Dispatch handlers for a message.

        Handles the retrieval and calling of handlers and the error state of no
        handlers being found.

        Returns:
            The created concurrent.futures.Futures in which the handler(s) are
            being run or, for high-priority handlers, the _PriorityHandler which
            the execution is tied to.
        """
        try:
            logger.debug(
                f"Attempting to dispatch handlers for"
                f" {message.message_type} from {message.sender}"
                f" in mailbox {self.mailbox_id}"
            )
            return await self._handle_message(message, handler_dispatch_loop, timeout)
        except MessageHandlerNotFoundError:
            self._default_handler(message)
            return []
        except MessageTypeSpecificHandlerNotFoundError as exc:
            logger.warning(
                f"No specific handler could be found for message "
                f"("
                f"type: {message.message_type}; "
                f"sender {message.sender}; "
                f"recipient {message.recipient}"
                f"). "
                f"Message was passed to universal handlers only."
            )
            return exc.universal_dispatches

    async def _handle_message(
        self,
        message: _BitfountMessage,
        handler_dispatch_loop: AbstractEventLoop,
        timeout: int = HANDLER_RETRIEVAL_TIMEOUT,
    ) -> List[Union[ConcurrentFuture, _PriorityHandler]]:
        """Finds and runs handler(s) for the supplied message.

        The lower-priority handler(s) (whether async or not) are run within an
        asyncio event loop (the handler_dispatch_loop) to avoid blocking
        listen_for_messages().

        High-priority handlers are run directly (although in actuality in a thread)
        to avoid a blocking async task from stopping them being executed.

        Args:
            message: The message to handle.
            handler_dispatch_loop: The event loop to dispatch the handler execution
                to. Generally this will be the event loop in the main thread.
            timeout: The time to allow handlers to be registered for dispatch.

        Returns:
            The created concurrent.futures.Futures in which the handler(s) are
            being run or, for high-priority handlers, the _PriorityHandler which
            the execution is tied to.

        Raises:
            MessageHandlerNotFoundError: If no handler is registered for message
                type and one is not registered within the timeout.
            MessageTypeSpecificHandlerNotFoundError: If only universal handlers
                are found for the message type.
        """
        message_type: _BitfountMessageType = message.message_type
        awaitables: List[Union[ConcurrentFuture, _PriorityHandler]] = []
        handlers_specific_or_universal: List[bool] = []

        # We create an async wrapper around any low-priority handlers regardless
        # of if it's an async function or not to allow us to run it in the background
        # as a Task. This also allows us to access the asyncio.Locks.
        #
        # We need to use a separate method to return the wrapper (rather than defining
        # it in the for-loop) due to the late-binding of closures in Python.
        def _get_run_handler_wrapper(
            handler: NonPriorityHandler,
        ) -> Callable[[], Coroutine[None, None, None]]:
            async def _run_handler_wrapper() -> None:
                # Only a single handler instance (i.e. a call to a specific handler
                # function) should be running at a given time; this helps avoid
                # conflicting use of shared resources and ensures that tasks we
                # _want_ to block (such as worker running) do so.
                # TODO: [BIT-1048] Revisit this and decide if _all_ messages need to
                #       be exclusive in this way.
                message_lock: Optional[AsyncContextManager]
                try:
                    message_lock = self._handlers.get_lock(message_type, handler)
                except KeyError:
                    message_lock = None

                if message_lock is None:
                    message_lock = asyncnullcontext()

                async with message_lock:
                    # As this supports both sync and async handlers we need to
                    # process the result (which should be None, but could be a
                    # Coroutine returning None). As such, we comfortably call the
                    # handler and then simply await the result if needed.
                    logger.debug(
                        f"Running low-priority handler for {message.message_type.name}"
                        f" message from {message.sender} in mailbox {self.mailbox_id}"
                    )
                    result = handler(message)
                    if inspect.isawaitable(result):
                        # Mypy needs some assurance, despite the above check
                        result = cast(Awaitable[None], result)
                        await result

            return _run_handler_wrapper

        # We retrieve and dispatch handlers up to the timeout limit from the async
        # generator. Any handlers already registered or that are registered in this
        # timeout window will be dispatched.
        async for handler, is_type_specific in self._retrieve_handlers(
            message_type, timeout
        ):
            logger.debug(
                f"Handler retrieved for {message_type.name}"
                f" in mailbox {self.mailbox_id}"
            )
            handlers_specific_or_universal.append(is_type_specific)

            handler_callable: Handler
            if not isinstance(handler, _TemporaryHandler):
                handler_callable = handler
            else:
                handler_callable = handler.handler

            if not isinstance(handler_callable, _PriorityHandler):
                # For non-priority handlers we farm them out to the handler dispatch
                # event loop. We return the concurrent Future instance so the task
                # can be monitored later.
                wrapped_handler = _get_run_handler_wrapper(handler_callable)
                logger.debug(
                    f"Dispatching low priority handler for {message_type.name}"
                    f" to event loop (id={id(handler_dispatch_loop)})"
                    f" in mailbox {self.mailbox_id}"
                )
                fut = asyncio.run_coroutine_threadsafe(
                    wrapped_handler(), handler_dispatch_loop
                )
                awaitables.append(fut)
            else:
                # The handler register locks aren't used for _PriorityHandler as
                # they get run in other threads and asyncio.Locks aren't thread
                # safe. The _PriorityHandler itself enforces locking.
                logger.debug(
                    f"Dispatching high priority handler for {message_type.name}"
                    f" in mailbox {self.mailbox_id}"
                )
                handler_callable(message)
                awaitables.append(handler_callable)

            # As handler has been dispatched, we should delete it if it's a
            # temporary handler
            if isinstance(handler, _TemporaryHandler):
                logger.debug(
                    f"Temporary handler {handler} has been dispatched for"
                    f" {message_type.name} in mailbox {self.mailbox_id}."
                    f" Deleting handler."
                )
                self._delete_handler(message_type, handler)

        # If no handlers were dispatched at all, raise exception
        if len(handlers_specific_or_universal) == 0:
            raise MessageHandlerNotFoundError(
                f"No handlers for message type {message_type.name}"
                f" were found in mailbox {self.mailbox_id}"
            )

        # If only universal handlers were found, raise exception
        if not any(handlers_specific_or_universal):
            exc = MessageTypeSpecificHandlerNotFoundError(
                f"Only universal handlers were found for message type"
                f" {message.message_type} in mailbox {self.mailbox_id}"
            )
            exc.universal_dispatches = awaitables
            raise exc

        return awaitables

    async def _retrieve_handlers(
        self,
        message_type: _BitfountMessageType,
        timeout: int = HANDLER_RETRIEVAL_TIMEOUT,
    ) -> AsyncIterator[Tuple[_RegisterHandler, bool]]:
        """Yield the registered handler(s) for the given message type.

        Args:
            message_type: Message type to retrieve handler for.
            timeout: Number of seconds to wait before stopping handler retrieval.

        Yields:
            The handler(s) registered for the given message type within the timeout.
        """
        try:
            async with async_timeout.timeout(timeout):
                # We also pass the timeout through to
                # _retrieve_handlers_with_monitoring() as this allows us to put
                # bounds on how long the threads that wait on threading.Events can run.
                async for handler, is_type_specific in self._retrieve_handlers_with_monitoring(  # noqa: B950
                    message_type, timeout
                ):
                    yield handler, is_type_specific
        except asyncio.TimeoutError:
            # This is expected from the async_timeout.timeout() contextmanager
            return

    async def _retrieve_handlers_with_monitoring(
        self,
        message_type: _BitfountMessageType,
        timeout: float,
    ) -> AsyncIterator[Tuple[_RegisterHandler, bool]]:
        """Yield handler(s) for message type as they become available.

        Stops yielding when a handler specific to that message type is available.

        Yields:
            A tuple of the handler and a boolean indicating whether the handler
            was specifically registered for that message type or was a universal
            handler.
        """
        # We give a small amount of buffer to ensure the Event waiting doesn't
        # finish before the async_timeout.timeout() in the parent method call.
        timeout *= 1.1
        start_time = time.time()

        # Allows us to avoid yielding a handler twice whilst allowing us to not
        # have to extract _only_ new handlers.
        yielded_handlers: Set[_RegisterHandler] = set()

        # We create a monitor event before initial yielding to detect any handlers
        # that are added during that initial yielding.
        new_handler_added_monitor = threading.Event()
        logger.debug(
            f"Trying to add new handler monitor for {message_type.name}"
            f" in mailbox {self.mailbox_id}"
        )
        async with async_lock(
            self._new_handler_added_monitors_lock, "new_handler_added_monitor_lock"
        ):
            logger.debug(
                f"Added new handler monitor for {message_type.name}"
                f" in mailbox {self.mailbox_id}"
            )
            self._new_handler_added_monitors.add(new_handler_added_monitor)

        try:
            # Do an initial yield for any handlers that are already registered
            message_type_specific_handler_found = False
            handlers = self._get_relevant_handlers(message_type)
            if any([is_type_specific for handler, is_type_specific in handlers]):
                message_type_specific_handler_found = True

            for handler, is_type_specific in handlers:
                logger.debug(
                    f"Retrieved handler {handler} for {message_type.name}"
                    f" in mailbox {self.mailbox_id}"
                )
                if handler not in yielded_handlers:
                    yielded_handlers.add(handler)
                    logger.debug(
                        f"Yielding {'universal ' if not is_type_specific else ''}"
                        f"handler {handler} for {message_type.name}"
                        f" in mailbox {self.mailbox_id}"
                    )
                    yield handler, is_type_specific
                else:
                    logger.debug(
                        f"Handler {handler} already dispatched for"
                        f" {message_type.name} in mailbox {self.mailbox_id}"
                    )

            # If we've already dispatched a message specific handler, return immediately
            if message_type_specific_handler_found:
                return

            # Now wait on any new handler registrations
            while True:
                # Get remaining timeout based on start time and current time
                remaining_timeout = timeout - (time.time() - start_time)

                # This condition should never happen due to timeout buffer
                if remaining_timeout < 0:
                    logger.warning(
                        f"Timeout encountered whilst retrieving handlers"
                        f" for {message_type}."
                    )
                    return

                # Wait for a new handler to be registered. If one is already registered
                # since the last time this waiting was attempted, returns immediately.
                # As self._new_handler_added is a `threading.Event` we run the
                # waiting in a separate thread to not block the event loop.
                monitor_set: bool = await await_threading_event(
                    new_handler_added_monitor,
                    event_name=f"new_handler_added_monitor_{message_type}",
                    timeout=remaining_timeout,
                )

                # monitor_set should always be true at this point due to timeout
                # buffer and await_threading_event only returning when Event.wait()
                # is achieved.
                if monitor_set:
                    logger.debug(
                        f"Detected new handler registration in {message_type.name}"
                        f" in mailbox {self.mailbox_id}"
                    )
                    # Reset monitor. No need to acquire lock here as even if
                    # _register_handler is mid-notify we'll either pick up the new
                    # handlers immediately or on the next loop.
                    new_handler_added_monitor.clear()

                    handlers = self._get_relevant_handlers(message_type)
                    if any(
                        [is_type_specific for handler, is_type_specific in handlers]
                    ):
                        message_type_specific_handler_found = True

                    for handler, is_type_specific in handlers:
                        logger.debug(
                            f"Retrieved handler {handler} for {message_type.name}"
                            f" in mailbox {self.mailbox_id}"
                        )
                        if handler not in yielded_handlers:
                            yielded_handlers.add(handler)
                            logger.debug(
                                f"Yielding "
                                f"{'universal ' if not is_type_specific else ''}"
                                f"handler {handler} for {message_type.name}"
                                f" in mailbox {self.mailbox_id}"
                            )
                            yield handler, is_type_specific
                        else:
                            logger.debug(
                                f"Handler {handler} already dispatched "
                                f"for {message_type.name} in mailbox {self.mailbox_id}"
                            )

                    # If we've now dispatched a message specific handler,
                    # return immediately
                    if message_type_specific_handler_found:
                        return
                else:
                    logger.warning(
                        f"Timeout encountered whilst waiting for notification"
                        f" of new handler registrations for {message_type}."
                    )
                    return
        finally:
            async with async_lock(
                self._new_handler_added_monitors_lock, "new_handler_added_monitor_lock"
            ):
                logger.debug(
                    f"Removing new handler monitor for {message_type.name}"
                    f" in mailbox {self.mailbox_id}"
                )
                self._new_handler_added_monitors.remove(new_handler_added_monitor)

                # Set so any threads that are waiting on it are done; this is only
                # used in this method so doesn't matter that we've set it even if
                # it is not actually a "new handler added" event.
                new_handler_added_monitor.set()

    def _get_relevant_handlers(
        self, message_type: _BitfountMessageType
    ) -> List[Tuple[_RegisterHandler, bool]]:
        """Retrieve all relevant handlers for a given message type.

        Includes both specific handlers and universal handlers.
        """
        handlers: List[Tuple[_RegisterHandler, bool]] = []

        # Get any existing handlers for that specific message type
        handlers.extend(
            (handler, True) for handler in self._handlers.get_handlers(message_type)
        )

        # Get any existing universal handlers
        handlers.extend(
            (handler, False) for handler in self._handlers.get_handlers(ANY_MESSAGE)
        )

        return handlers

    def _register_handler(
        self, message_type: _ExtendedMessageTypes, handler: _RegisterHandler
    ) -> None:
        """Registers a handler for a specific message type."""
        self._handlers.register_handler(message_type, handler)

        # Note we've added a new handler, to allow backed off handler retrieval
        # to know.
        start_wait_time = time.time()
        with self._new_handler_added_monitors_lock:
            if _BITFOUNT_MULTITHREADING_DEBUG:
                logger.debug(
                    f"Was waiting on _new_handler_added_monitors_lock in"
                    f" _register_handler() for {time.time()-start_wait_time:.4f}"
                    f" seconds."
                )

            if self._new_handler_added_monitors:
                message_type_name = (
                    message_type.name
                    if isinstance(message_type, _BitfountMessageType)
                    else message_type
                )
                for monitor in self._new_handler_added_monitors:
                    monitor.set()

                if _BITFOUNT_MULTITHREADING_DEBUG:
                    logger.debug(
                        f"Notified {len(self._new_handler_added_monitors)} handlers"
                        f" in mailbox {self.mailbox_id}"
                        f" due to {message_type_name} handler {handler}"
                    )

        if _BITFOUNT_MULTITHREADING_DEBUG:
            logger.debug(
                "Released _new_handler_added_monitors_lock in _register_handler()"
            )

    def register_handler(
        self,
        message_type: _BitfountMessageType,
        handler: Handler,
        high_priority: bool = False,
    ) -> Handler:
        """Registers a handler for a specific message type.

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered handler, which may not be the same as the supplied handler.
        """
        handler = self._process_handler(handler, high_priority)

        logger.debug(
            f"Registering handler {handler} for message type {message_type}"
            f" in mailbox {self.mailbox_id}"
        )
        self._register_handler(message_type, handler)

        return handler

    def register_universal_handler(
        self, handler: Handler, high_priority: bool = False
    ) -> Handler:
        """Registers a universal handler, that will also be called for all messages.

        :::caution

        Universal handlers are run IN ADDITION to any other handlers for that message
        type. If no non-universal handlers are registered for that message type,
        the default handler will be used instead.

        :::

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered handler, which may not be the same as the supplied handler.
        """
        handler = self._process_handler(handler, high_priority)

        logger.debug(
            f"Registering universal handler {handler} in mailbox {self.mailbox_id}"
        )
        self._register_handler(ANY_MESSAGE, handler)

        return handler

    def register_temp_handler(
        self,
        message_type: _BitfountMessageType,
        handler: Handler,
        high_priority: bool = False,
    ) -> _TemporaryHandler:
        """Registers a handler that will be deleted after it is called.

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered temporary handler.
        """
        handler = self._process_handler(handler, high_priority)
        temp_handler = self._make_temp_handler(handler)

        logger.debug(
            f"Registering temporary handler {handler} for message type {message_type}"
            f" in mailbox {self.mailbox_id}"
        )
        self._register_handler(message_type, temp_handler)

        return temp_handler

    def register_temp_universal_handler(
        self, handler: Handler, high_priority: bool = False
    ) -> _TemporaryHandler:
        """Registers a universal handler that will be deleted after it is called.

        :::caution

        Universal handlers are run IN ADDITION to any other handlers for that message
        type. If no non-universal handlers are registered for that message type,
        the default handler will be used instead.

        :::

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered temporary handler.
        """
        handler = self._process_handler(handler, high_priority)
        temp_handler = self._make_temp_handler(handler)

        logger.debug(
            f"Registering temporary universal handler {handler}"
            f" in mailbox {self.mailbox_id}"
        )
        self._register_handler(ANY_MESSAGE, temp_handler)

        return temp_handler

    @staticmethod
    def _make_temp_handler(handler: Handler) -> _TemporaryHandler:
        """Wraps handler so that it will be deleted after it is called."""
        return _TemporaryHandler(handler)

    def _delete_handler(
        self, message_type: _BitfountMessageType, handler: _RegisterHandler
    ) -> None:
        """Delete a handler from the registry."""
        self._handlers.delete_handler(message_type, handler)

    def delete_handler(
        self,
        message_type: _BitfountMessageType,
        handler: Optional[_RegisterHandler] = None,
    ) -> None:
        """Deletes a handler associated with the message type.

        If a specific handler is not provided, deletes all handlers for that
        message type.
        """
        if not handler:
            warnings.warn(
                "In future versions, delete_handler will require a specific handler"
                " instance to be provided. Please switch to delete_all_handlers"
                " instead to maintain previous functionality.",
                DeprecationWarning,
            )
            self.delete_all_handlers(message_type)
        else:
            logger.debug(
                f"Deleting handler for message type: {message_type}"
                f" in mailbox {self.mailbox_id}"
            )
            self._delete_handler(message_type, handler)

    def delete_universal_handler(self, handler: _RegisterHandler) -> None:
        """Deletes a specific universal handler."""
        logger.debug(
            f"Deleting universal handler {handler} in mailbox {self.mailbox_id}"
        )
        self._handlers.delete_handler(ANY_MESSAGE, handler)

    def delete_all_handlers(self, message_type: _BitfountMessageType) -> None:
        """Deletes all handlers for a specific message type."""
        logger.debug(
            f"Deleting all handlers for message type: {message_type}"
            f" in mailbox {self.mailbox_id}"
        )
        self._handlers.delete_all_handlers(message_type)

    def delete_all_universal_handlers(self) -> None:
        """Deletes all universal handlers."""
        logger.debug(f"Deleting all universal handlers in mailbox {self.mailbox_id}")
        self._handlers.delete_all_handlers(ANY_MESSAGE)

    @staticmethod
    def _default_handler(message: _BitfountMessage) -> None:
        """Simple default handler that logs the message details.

        If this is called it is because we have received a message type that we
        were not expecting and do not know how to handle. We log out pertinent
        (non-private) details.
        """
        logger.error(
            f"Received unexpected message "
            f"("
            f"type: {message.message_type}; "
            f"sender {message.sender}; "
            f"recipient {message.recipient}"
            f"). "
            f"Message was not handled."
        )

    @staticmethod
    def _process_handler(handler: Handler, high_priority: bool) -> Handler:
        """Process the supplied handler given handler configuration.

        Handles conversion into high priority handlers as needed and also does
        type-checking for supported handler types.

        Return:
            The processed (and hence potentially changed) handler.
        """
        if high_priority:
            if isinstance(handler, _PriorityHandler):
                # Already correct, just return
                return handler
            else:
                logger.debug(f"Converting handler {handler} to high priority handler.")
                return _PriorityHandler(handler)
        else:
            if isinstance(handler, _PriorityHandler):
                logger.warning(
                    "A priority handler has been provided but high_priority not set."
                    " Treating as though high_priority is True."
                )
                return handler
            else:
                return handler


async def _send_aes_encrypted_message(
    message: Any,
    aes_encryption_key: bytes,
    message_service: _MessageService,
    **kwargs: Any,
) -> None:
    """Packs message, encrypts it and sends it.

    Args:
        message: The message to be sent. Must support serialisation via msgpack.
        aes_encryption_key: Key used to encrypt message.
        message_service: The MessageService used to send the message.
        **kwargs: Keyword arguments passed to BitfountMessage constructor.
    """
    body = msgpack.dumps(message, default=msgpackext_encode)
    message_body = _MessageEncryption.encrypt_outgoing_message(body, aes_encryption_key)

    await message_service.send_message(
        _BitfountMessage(body=message_body, **kwargs),
        already_packed=True,
    )


def _decrypt_aes_message(message: _BitfountMessage, aes_encryption_key: bytes) -> Any:
    """Decrypt AES encrypted message.

    Args:
        message: Encrypted message to decrypt.
        aes_encryption_key: The AES key to use to decrypt.

    Returns:
        Decrypted message body.
    """
    body = _MessageEncryption.decrypt_incoming_message(message.body, aes_encryption_key)
    return msgpack.loads(body, ext_hook=msgpackext_decode)


# Type variable for return type below.
_R = TypeVar("_R")


async def _run_func_and_listen_to_mailbox(
    run_func: Coroutine[None, None, _R],
    mailbox: _BaseMailbox,
) -> _R:
    """Runs an async function and listens for messages simultaneously.

    This function allows any exceptions that occur in the run function to be properly
    propagated to the calling code whilst ensuring that the listener or run function
    are correctly shutdown in such a situation.

    It also ensures that the mailbox listener is not run for longer than the lifetime
    of run_func.

    Args:
        run_func: The function to run that will be needing received messages.
        mailbox: The mailbox to use to listen for messages.

    Returns:
         The return value of run_func.
    """
    with mailbox.listen() as mailbox_listener:
        loop = asyncio.get_running_loop()
        executor = ThreadPoolExecutor(
            1, thread_name_prefix="run_func_and_listen_executor"
        )
        try:
            # Create a separate task for the wrapped function call
            run_task: Task[_R] = asyncio.create_task(run_func)

            # Create future to monitor mailbox listener
            fut: AsyncFuture[None] = loop.run_in_executor(
                executor, mailbox_listener.join
            )

            aws: List[Awaitable] = [run_task, fut]
            done, pending = await asyncio.wait(aws, return_when=FIRST_COMPLETED)

            if run_task in done:
                # Task has completed, return the result (or raise an exception)
                logger.debug(f"Task {run_task} is done")
                return run_task.result()
            else:  # run_task not in done
                # The mailbox listener has finished prematurely (likely due to an
                # error).
                exc = fut.exception()
                error_msg = (
                    f"Mailbox has finished listening before the target task"
                    f" ({run_task}) finished running."
                )

                if exc:
                    # If an exception occurred in the mailbox listener, log information
                    # about it and then throw the exception into the run_task to handle
                    error_msg += (
                        f" Exception was: {exc}. This has been passed to the task"
                        f" to handle."
                    )
                    logger.error(error_msg)
                    logger.exception(exc)

                    # Pass to task to handle then await on task
                    cast(Coroutine, run_task.get_coro()).throw(exc)
                    return await run_task
                else:
                    # Otherwise, log that something went wrong and wait on the run_task
                    # to "finish"
                    logger.error(error_msg)
                    return await run_task
        finally:
            executor.shutdown(wait=False)



"""Algorithm to evaluate a model on remote data."""
from __future__ import annotations

from typing import Any, Dict, List, Mapping, cast

import numpy as np

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithmFactory,
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.hub.api import BitfountHub
from bitfount.utils import delegates

logger = _get_federated_logger(__name__)


class _ModellerSide(_BaseModellerModelAlgorithm):
    """Modeller side of the ModelInference algorithm."""

    def run(
        self, predictions: Mapping[str, List[np.ndarray]]
    ) -> Dict[str, List[np.ndarray]]:
        """Simply returns predictions."""
        return dict(predictions)


class _WorkerSide(_BaseWorkerModelAlgorithm):
    """Worker side of the ModelInference algorithm."""

    def run(self, data: BaseSource) -> List[np.ndarray]:
        """Runs evaluation and returns metrics."""
        return cast(List[np.ndarray], self.model.predict(data))


@delegates()
class ModelInference(_BaseModelAlgorithmFactory):
    """Algorithm for running inference on a model and returning the predictions.

    :::danger

    This algorithm could potentially return the data unfiltered so should only be used
    when the other party is trusted.

    :::

    Args:
        model: The model to infer on remote data.

    Attributes:
        model: The model to infer on remote data.
    """

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the ModelInference algorithm."""
        model = self._get_model_from_reference()
        return _ModellerSide(model=model, **kwargs)

    def worker(self, hub: BitfountHub, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the ModelInference algorithm.

        Args:
            hub: `BitfountHub` object to use for communication with the hub.
        """
        model = self._get_model_from_reference(hub=hub)
        return _WorkerSide(model=model, **kwargs)



"""Tests tutorials individually."""
import importlib
import logging
import multiprocessing
import os
from pathlib import Path
import shutil
import time
from typing import Dict, Final, Generator, Optional

from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempPathFactory
import jupytext
import nbformat
from nbformat import NotebookNode
import pytest
from pytest import fixture

from bitfount.config import _PRODUCTION_ENVIRONMENT, _STAGING_ENVIRONMENT
from tests.integration.tutorials.notebook_logging import (
    _LOG_QUEUE_AUTH_KEY,
    _LOG_QUEUE_PORT,
    LogQueueManager,
)
from tests.integration.utils import (
    _patch_authentication_code,
    get_patched_authentication_code_nb,
)
from tests.utils.helper import (
    tutorial_2_test,
    tutorial_4_test,
    tutorial_6_test,
    tutorial_7_test,
    tutorial_8_test,
    tutorial_9_test,
    tutorial_11_test,
    tutorial_test,
)

# User details for running tests
logger = logging.getLogger(__name__)

PASSWORD_ENVVAR: Final[str] = "BF_USER_PASSWORD"
ENVIRONMENT_ENVVAR: Final[str] = "BITFOUNT_ENVIRONMENT"
STAGING_USERNAME: Final[str] = "e2e_modeller"
PRODUCTION_USERNAME: Final[str] = "bitfount-rel"


MODELLER_TIMEOUT_PERIOD: Final[int] = 5 * 60  # 5 minutes
IMAGE_MODELLER_TIMEOUT_PERIOD: Final[int] = 12 * 60  # 12 minutes
MULTI_POD_MODELLER_TIMEOUT_PERIOD: Final[int] = 12 * 60  # 12 minutes
POD_TEST_TIMEOUT_PERIOD: Final[int] = 3 * 60  # 3 minutes

# This is the maximum time that a _background_ Pod will be running for.
# All tasks should comfortably be run in this time.
TUTORIALS_PATH: str = "tutorials"
POD_TIMEOUT_PERIOD: Final[int] = 45 * 60  # 45 minutes
POD_TUTORIALS_DIR = "Connecting Data & Creating Pods"
BASIC_DATA_SCIENCE_TUTORIALS_DIR = "Running Basic Data Science Tasks"
ADVANCED_DATA_SCIENCE_TUTORIALS_DIR = "Advanced Data Science Tasks"
PRIVACY_PRESERVING_TUTORIALS_DIR = "Privacy-Preserving Techniques"
TUTORIALS: Final[Dict[int, str]] = {
    1: f"{TUTORIALS_PATH}/{POD_TUTORIALS_DIR}/running_a_pod",
    2: f"{TUTORIALS_PATH}/{BASIC_DATA_SCIENCE_TUTORIALS_DIR}/querying_and_training_a_model",  # noqa: B950
    3: f"{TUTORIALS_PATH}/{POD_TUTORIALS_DIR}/running_a_pod_using_yaml",
    4: f"{TUTORIALS_PATH}/{BASIC_DATA_SCIENCE_TUTORIALS_DIR}/training_a_model_on_two_pods",  # noqa: B950
    5: f"{TUTORIALS_PATH}/{POD_TUTORIALS_DIR}/running_an_image_data_pod",
    6: f"{TUTORIALS_PATH}/{BASIC_DATA_SCIENCE_TUTORIALS_DIR}/training_on_images",
    7: f"{TUTORIALS_PATH}/{ADVANCED_DATA_SCIENCE_TUTORIALS_DIR}/training_a_custom_model",  # noqa: B950
    8: f"{TUTORIALS_PATH}/{ADVANCED_DATA_SCIENCE_TUTORIALS_DIR}/using_pretrained_models",  # noqa: B950
    9: f"{TUTORIALS_PATH}/{PRIVACY_PRESERVING_TUTORIALS_DIR}/privacy_preserving_techniques",  # noqa: B950
    10: f"{TUTORIALS_PATH}/{POD_TUTORIALS_DIR}/running_a_segmentation_data_pod",
    11: f"{TUTORIALS_PATH}/{ADVANCED_DATA_SCIENCE_TUTORIALS_DIR}/training_a_custom_segmentation_model",  # noqa: B950
}

# Constants representing the names of the pods as in the tutorials
CENSUS_INCOME_POD_NAME: Final[str] = "census-income-demo"
CENSUS_INCOME_YAML_POD_NAME: Final[str] = "census-income-yaml-demo"
MNIST_POD_NAME: Final[str] = "mnist-demo"
SEGMENTATION_POD_NAME: Final[str] = "segmentation-data-demo"

# Path details
BITFOUNT_ROOT: Final[Path] = Path(__file__).parent.parent.parent.parent
TUTORIALS_ROOT: Final[Path] = BITFOUNT_ROOT / "tutorials"
TEST_SCHEMAS_PATH: Final[Path] = (
    BITFOUNT_ROOT / "tests" / "integration" / "resources" / "schemas"
)


@fixture
def username() -> str:
    """Name of user hosting pod."""
    return "test_username"


@fixture(scope="module")
def pod_suffix() -> str:
    """Generate a unique suffix for all pods in this test run."""
    return str(int(time.time()))  # unix timestamp


@pytest.fixture(scope="module")
def monkeypatch_module_scope() -> Generator[MonkeyPatch, None, None]:
    """Module-scoped monkeypatch."""
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


@fixture(autouse=True, scope="module")
def set_bitfount_environment(
    bitfount_env: str, monkeypatch_module_scope: MonkeyPatch
) -> None:
    """Monkeypatches Bitfount environment for tutorial tests.

    Guarantees that the envvar will be restored to the original value when
    tests finish.
    """
    monkeypatch_module_scope.setenv(ENVIRONMENT_ENVVAR, bitfount_env)


@fixture(scope="module")
def bitfount_env() -> str:
    """Determines the Bitfount environment to run the tests against.

    Defaults to staging if not explicitly set.
    """
    env = os.getenv(ENVIRONMENT_ENVVAR, _STAGING_ENVIRONMENT)
    logging.info(f'Using the "{env}" environment for these tutorial tests.')
    return env


@fixture(scope="module")
def authentication_user(bitfount_env: str) -> str:
    """Returns the username to use for authentication."""
    if bitfount_env == _STAGING_ENVIRONMENT:
        user = STAGING_USERNAME
    elif bitfount_env == _PRODUCTION_ENVIRONMENT:
        user = PRODUCTION_USERNAME
    else:
        raise ValueError(
            f"No authentication user defined"
            f' for Bitfount environment "{bitfount_env}"'
        )

    logging.info(f'Authentication user "{user}" will be used for these tests.')
    return user


@fixture(scope="module")
def authentication_password() -> str:
    """Returns the password for logging in as the target user.

    This will be supplied as the BF_USER_PASSWORD environment variable.
    """
    if not (pswd := os.getenv(PASSWORD_ENVVAR)):
        raise ValueError("No user password provided, unable to proceed.")
    else:
        return pswd


@fixture(scope="module")
def token_dir(tmp_path_factory: TempPathFactory) -> Path:
    """Temporary directory for tokens."""
    return tmp_path_factory.mktemp(".bitfount", numbered=False) / "bitfount_tokens"


@fixture(scope="function")
def patched_authentication_code(
    authentication_password: str, authentication_user: str, token_dir: Path
) -> str:
    """Returns code that patches bitfount session and OIDC authentication.

    This enables them to be run in a headless environment and still perform
    the OAuth logins required.
    """
    return get_patched_authentication_code_nb(
        authentication_user, authentication_password, token_dir
    )


@fixture(scope="module")
def census_income_pod_name(pod_suffix: str) -> str:
    """Name of the census income pod."""
    pod_name = f"{CENSUS_INCOME_POD_NAME}-{pod_suffix}"
    logger.info(f"Census income pod name: {pod_name}")
    return pod_name


@fixture(scope="module")
def census_income_yaml_pod_name(pod_suffix: str) -> str:
    """Name of the census income pod."""
    pod_name = f"{CENSUS_INCOME_YAML_POD_NAME}-{pod_suffix}"
    logger.info(f"Census income yaml pod name: {pod_name}")
    return pod_name


def _replace_in_notebook_source_code(
    query: str, replacement: str, notebook: NotebookNode
) -> NotebookNode:
    """Replaces all occurrences of `query` with `replacement` in `notebook` source code.

    Args:
        query: code to replace
        replacement: replacement code
        notebook: notebook to have code replaced

    Returns:
        The same notebook with updated source code.
    """
    loggable_query = query.replace("\n", "\\n")
    loggable_replacement = replacement.replace("\n", "\\n")
    logger.info(f'Replacing "{loggable_query}" with "{loggable_replacement}" ')

    changes_made = False
    for i, cell in enumerate(notebook["cells"]):
        orig_source_code: str = cell["source"]
        if cell["cell_type"] == "code" and query in orig_source_code:
            # Mark if changes have been made so we know that things have been
            # applied correctly; every replace should cause at least one change.
            # We know it will be applied because of the `query in` check above.
            changes_made = True
            new_source_code = orig_source_code.replace(query, replacement)
            notebook["cells"][i]["source"] = new_source_code

    if not changes_made:
        raise AssertionError(
            f'No code was replaced when trying to replace "{loggable_query}" '
            f'with "{loggable_replacement}"'
        )
    return notebook


@fixture(scope="module")
def extra_imports() -> str:
    """Extra imports needed for testing."""
    return (
        "from bitfount.federated.pod_keys_setup import _get_pod_keys\n"
        "from pathlib import Path\n"
        "from typing import Tuple, Any\n"
        "import unittest"
    )


@fixture(scope="module")
def change_dir_code() -> str:
    """Changes the directory of the jupyter notebook execution to `tutorials`."""
    return ""


@fixture(scope="module")
def log_queue_manager() -> Generator[LogQueueManager, None, None]:
    """Hosts and starts a LogQueueManager instance for logging messages."""
    with LogQueueManager(
        address=("", _LOG_QUEUE_PORT), authkey=_LOG_QUEUE_AUTH_KEY
    ) as manager:
        logger.info("LogQueueManager started")
        yield manager


@fixture
def keystore_path(tmp_path_factory: TempPathFactory) -> Path:
    """Temporary file for key storage."""
    return tmp_path_factory.mktemp("key_store") / "known_workers.yml"


@fixture(scope="module")
def mnist_pod_name(pod_suffix: str) -> str:
    """Name of the MNIST pod."""
    mnist_pod_name = f"{MNIST_POD_NAME}-{pod_suffix}"
    logger.info(f"MNIST pod name: {mnist_pod_name}")
    return mnist_pod_name


@fixture(scope="module")
def seg_pod_name(pod_suffix: str) -> str:
    """Name of the Segmentation pod."""
    seg_pod_name = f"{SEGMENTATION_POD_NAME}-{pod_suffix}"
    logger.info(f"Segmentation pod name: {seg_pod_name}")
    return seg_pod_name


def start_census_income_pod(
    authentication_password: str,
    census_income_pod_name: str,
    authentication_user: str,
    extra_imports: str,
    token_dir: Path,
    keystore_path: Path,
    census_income_yaml_pod_name: Optional[str] = None,
) -> None:
    """Tutorial 1 pod - tested in multiple tests."""
    # Change the notebook from markdown and replace pod name
    ntbk = jupytext.read(f"{TUTORIALS[1]}.md")
    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_POD_NAME, census_income_pod_name, ntbk
    )
    if census_income_yaml_pod_name is not None:
        ntbk = _replace_in_notebook_source_code(
            CENSUS_INCOME_YAML_POD_NAME, census_income_yaml_pod_name, ntbk
        )

    ntbk = _replace_in_notebook_source_code(
        "pod = Pod(",
        f'{extra_imports}\npod_keys=_get_pod_keys( Path("{str(token_dir)}")/"pods" / "{census_income_pod_name}")\npod = Pod(pod_keys=pod_keys,',  # noqa: B950
        ntbk,
    )

    ntbk = _replace_in_notebook_source_code(
        "data_config=PodDataConfig(",
        f'username="{authentication_user}",\ndata_config=PodDataConfig(',
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    with open(f"{TUTORIALS[1]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    # Execute the notebook
    importlib.import_module(f"ipynb.fs.full.{TUTORIALS[1]}")


@tutorial_test
@tutorial_2_test
def test_tutorial_2(
    authentication_password: str,
    authentication_user: str,
    census_income_pod_name: str,
    extra_imports: str,
    keystore_path: Path,
    token_dir: Path,
) -> None:
    """Test for Tutorial 2 - querying_and_training_a_model.md."""
    # Set up pod
    pod_process = multiprocessing.Process(
        target=start_census_income_pod,
        name="Pod_Runner",
        args=(
            authentication_password,
            census_income_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
            keystore_path,
        ),
    )
    pod_process.start()
    time.sleep(40)  # give the pods time to spin up
    assert pod_process.is_alive()

    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[2]}.md")

    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_POD_NAME, census_income_pod_name, ntbk
    )
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/census_income_schema.yaml')\n"  # noqa: B950
        f"schema.tables[0].name = '{census_income_pod_name}'",
        ntbk,
    )
    with open(f"{TUTORIALS[2]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    # Execute the notebook
    notebook = importlib.import_module(f"ipynb.fs.full.{TUTORIALS[2]}")
    # Example for testing different vars in the notebooks
    assert (
        "occupation"
        in notebook.query_result[
            f"{authentication_user}/{census_income_pod_name}"
        ].keys()
    )

    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()
    os.remove(f"{TUTORIALS[2]}.ipynb")
    os.remove(f"{TUTORIALS[1]}.ipynb")


def start_yaml_census_income(
    authentication_password: str,
    census_income_yaml_pod_name: str,
    authentication_user: str,
    token_dir: Path,
    keystore_path: Path,
    census_income_pod_name: str,
) -> None:
    """Tutorial 3, tested in the tutorial 4 test."""
    # Try changing working directory if not already changed
    ntbk = jupytext.read(f"{TUTORIALS[3]}.md")
    ntbk = _replace_in_notebook_source_code(
        f"name: {CENSUS_INCOME_YAML_POD_NAME}\ndatasource: CSVSource",
        f"username: {authentication_user}\nname: {CENSUS_INCOME_YAML_POD_NAME}\ndatasource: CSVSource",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_YAML_POD_NAME, census_income_yaml_pod_name, ntbk
    )
    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_POD_NAME, census_income_pod_name, ntbk
    )

    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    with open(f"{TUTORIALS[3]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    importlib.import_module(f"ipynb.fs.full.{TUTORIALS[3]}")


@tutorial_test
@tutorial_4_test
def test_tutorial_4(
    authentication_password: str,
    authentication_user: str,
    census_income_pod_name: str,
    census_income_yaml_pod_name: str,
    extra_imports: str,
    keystore_path: Path,
    token_dir: Path,
) -> None:
    """Test for Tutorial 4 - training_a_model_on_two_pods.md."""
    # Set up the census-income pod
    pod_process = multiprocessing.Process(
        target=start_census_income_pod,
        name="Pod_Runner",
        args=(
            authentication_password,
            census_income_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
            keystore_path,
            census_income_yaml_pod_name,
        ),
    )
    pod_process.start()
    # Set up census-income-yaml pod
    yaml_pod_process = multiprocessing.Process(
        target=start_yaml_census_income,
        name="YAML_Pod_Runner",
        args=(
            authentication_password,
            census_income_yaml_pod_name,
            authentication_user,
            token_dir,
            keystore_path,
            census_income_pod_name,
        ),
    )
    yaml_pod_process.start()
    time.sleep(40)  # give the pods time to spin up
    assert pod_process.is_alive()
    assert yaml_pod_process.is_alive()

    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[4]}.md")
    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_POD_NAME, census_income_pod_name, ntbk
    )
    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_YAML_POD_NAME, census_income_yaml_pod_name, ntbk
    )
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = combine_pod_schemas([first_pod_identifier, second_pod_identifier])",  # noqa: B950
        f"schema = BitfountSchema.load_from_file('{str(TEST_SCHEMAS_PATH)}/census_income_schema.yaml')\n"  # noqa: B950
        "import copy;schema.tables.append(copy.deepcopy(schema.tables[0]))\n"  # noqa: B950
        f"schema.tables[0].name = '{census_income_pod_name}';schema.tables[1].name = '{census_income_yaml_pod_name}'",  # noqa: B950
        ntbk,
    )
    with open(f"{TUTORIALS[4]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    # Execute the notebook
    notebook = importlib.import_module(  # noqa: B950, F841 # this will be used when updating the tests
        f"ipynb.fs.full.{TUTORIALS[4]}"
    )

    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()
    yaml_pod_process.terminate()
    # terminate processes
    os.remove(f"{TUTORIALS[3]}.ipynb")
    os.remove(f"{TUTORIALS[1]}.ipynb")
    os.remove(f"{TUTORIALS[4]}.ipynb")


def start_mnist_pod(
    authentication_password: str,
    mnist_pod_name: str,
    authentication_user: str,
    extra_imports: str,
    token_dir: Path,
) -> None:
    """Tutorial 5 - tested in the test for tutorial 6."""
    ntbk = jupytext.read(f"{TUTORIALS[5]}.md")
    old_pod_name = MNIST_POD_NAME
    replacement_pod_name = mnist_pod_name
    ntbk = _replace_in_notebook_source_code(old_pod_name, replacement_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "pod = Pod(",
        f'{extra_imports}\npod_keys=_get_pod_keys( Path("{str(token_dir)}")/"pods" / "{replacement_pod_name}")\npod = Pod(pod_keys=pod_keys,',  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "%%writefile extract.py",
        "",
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "%run extract.py",
        "",
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nimport subprocess\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "data_config=image_data_config,",
        f'username="{authentication_user}",\ndata_config=image_data_config,',
        ntbk,
    )
    with open(f"{TUTORIALS[5]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    importlib.import_module(f"ipynb.fs.full.{TUTORIALS[5]}")


@tutorial_test
@tutorial_6_test
def test_tutorial_6(
    authentication_password: str,
    authentication_user: str,
    extra_imports: str,
    keystore_path: Path,
    mnist_pod_name: str,
    token_dir: Path,
) -> None:
    """Test for tutorial 6 - running_an_image_data_pod.md."""
    # Set up MNIST pod
    pod_process = multiprocessing.Process(
        target=start_mnist_pod,
        name="MNIST_Pod_Runner",
        args=(
            authentication_password,
            mnist_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
        ),
    )
    pod_process.start()
    time.sleep(40)  # give the pods time to spin up
    assert pod_process.is_alive()

    # Get the notebook
    ntbk = jupytext.read(f"{TUTORIALS[6]}.md")
    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = _replace_in_notebook_source_code(MNIST_POD_NAME, mnist_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/mnist_schema.yaml')\n"
        f"schema.tables[0].name = '{mnist_pod_name}'",
        ntbk,
    )
    with open(f"{TUTORIALS[6]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()

    # Execute the notebook
    notebook = importlib.import_module(  # noqa: B950, F841 # this will be used when updating the tests
        f"ipynb.fs.full.{TUTORIALS[6]}"
    )

    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()
    os.remove(f"{TUTORIALS[6]}.ipynb")
    os.remove(f"{TUTORIALS[5]}.ipynb")
    os.remove("mnist_labels.csv")
    os.remove("mnist_images.zip")
    shutil.rmtree("mnist_images")


@tutorial_test
@tutorial_7_test
def test_tutorial_7(
    authentication_password: str,
    authentication_user: str,
    census_income_pod_name: str,
    extra_imports: str,
    keystore_path: Path,
    token_dir: Path,
) -> None:
    """Test for tutorial 7 - training_a_custom_model.md."""
    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[7]}.md")
    old_pod_name = CENSUS_INCOME_POD_NAME
    replacement_pod_name = census_income_pod_name
    ntbk = _replace_in_notebook_source_code(old_pod_name, replacement_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nfrom MyCustomModel import MyCustomModel\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/census_income_schema.yaml')\n"  # noqa: B950
        f"schema.tables[0].name = '{census_income_pod_name}'",
        ntbk,
    )
    # Get the custom model from the notebook and save it separately
    for cell in ntbk.cells:
        if cell.cell_type == "code" and "class MyCustomModel" in cell.source:
            with open("MyCustomModel.py", mode="w", encoding="utf-8") as f:
                imports = """import torch\nfrom torch import nn as nn\nfrom torch.nn import functional as F\nfrom torchmetrics.functional import accuracy\nfrom bitfount import PyTorchBitfountModel, PyTorchClassifierMixIn\n"""  # noqa: B950
                model = imports + cell.source

                f.write(model)
            cell.source = ""
    with open(f"{TUTORIALS[7]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)
    # set up pod
    pod_process = multiprocessing.Process(
        target=start_census_income_pod,
        name="Pod_Runner",
        args=(
            authentication_password,
            census_income_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
            keystore_path,
        ),
    )
    pod_process.start()
    time.sleep(40)  # give the pods time to spin up
    assert pod_process.is_alive()
    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    # Execute the notebook
    notebook = importlib.import_module(  # noqa: B950, F841 # this will be used when updating the tests
        f"ipynb.fs.full.{TUTORIALS[7]}"
    )
    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()
    # cleanup
    os.remove(f"{TUTORIALS[1]}.ipynb")
    os.remove(f"{TUTORIALS[7]}.ipynb")
    os.remove("MyCustomModel.py")


@tutorial_test
@tutorial_8_test
def test_tutorial_8(
    authentication_password: str,
    authentication_user: str,
    census_income_pod_name: str,
    extra_imports: str,
    keystore_path: Path,
    token_dir: Path,
) -> None:
    """Tests tutorial 8 - using_pretrained_models.md."""
    # set up pod
    pod_process = multiprocessing.Process(
        target=start_census_income_pod,
        name="Pod_Runner",
        args=(
            authentication_password,
            census_income_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
            keystore_path,
        ),
    )
    pod_process.start()
    time.sleep(40)  # give the pods time to spin up
    assert pod_process.is_alive()

    # Set up the notebook from tutorial 2 as it is used here
    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[2]}.md")
    old_pod_name = CENSUS_INCOME_POD_NAME
    replacement_pod_name = census_income_pod_name
    ntbk = _replace_in_notebook_source_code(old_pod_name, replacement_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/census_income_schema.yaml')\n"  # noqa: B950
        f"schema.tables[0].name = '{census_income_pod_name}'",
        ntbk,
    )
    # We only need the model fitting and serialization from tutorial 2.
    for cell in ntbk.cells:
        if (
            cell.cell_type == "code"
            and ("query" in cell.source or "protocol" in cell.source)
            and "import" not in cell.source
        ):
            cell.source = ""
    with open(f"{TUTORIALS[2]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[8]}.md")
    ntbk = _replace_in_notebook_source_code(
        CENSUS_INCOME_POD_NAME, census_income_pod_name, ntbk
    )
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/census_income_schema.yaml')\n"  # noqa: B950
        f"schema.tables[0].name = '{census_income_pod_name}'",
        ntbk,
    )
    with open(f"{TUTORIALS[8]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()

    # Execute the notebooks
    # First, Tutorial 2
    importlib.import_module(f"ipynb.fs.full.{TUTORIALS[2]}")
    # Now, Tutorial 8
    notebook = importlib.import_module(  # noqa: B950, F841 # this will be used when updating the tests
        f"ipynb.fs.full.{TUTORIALS[8]}"
    )
    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()
    # cleanup
    os.remove(f"{TUTORIALS[1]}.ipynb")
    os.remove(f"{TUTORIALS[2]}.ipynb")
    os.remove(f"{TUTORIALS[8]}.ipynb")
    os.remove("training_a_model.pt")


@tutorial_test
@tutorial_9_test
def test_tutorial_9(
    authentication_password: str,
    authentication_user: str,
    census_income_pod_name: str,
    extra_imports: str,
    keystore_path: Path,
    token_dir: Path,
) -> None:
    """Test for Tutorial 9 - privacy_preserving_techniques.md."""
    # set up pod
    pod_process = multiprocessing.Process(
        target=start_census_income_pod,
        name="Pod_Runner",
        args=(
            authentication_password,
            census_income_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
            keystore_path,
        ),
    )
    pod_process.start()
    time.sleep(40)  # give the pods time to spin up
    assert pod_process.is_alive()

    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[9]}.md")
    old_pod_name = CENSUS_INCOME_POD_NAME
    replacement_pod_name = census_income_pod_name
    ntbk = _replace_in_notebook_source_code(old_pod_name, replacement_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/census_income_schema.yaml')\n"  # noqa: B950
        f"schema.tables[0].name = '{census_income_pod_name}'",
        ntbk,
    )
    with open(f"{TUTORIALS[9]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()

    # Execute the notebook
    notebook = importlib.import_module(  # noqa: B950, F841 # this will be used when updating the tests
        f"ipynb.fs.full.{TUTORIALS[9]}"
    )
    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()
    os.remove(f"{TUTORIALS[9]}.ipynb")
    os.remove(f"{TUTORIALS[1]}.ipynb")


def start_segmentation_pod(
    authentication_password: str,
    seg_pod_name: str,
    authentication_user: str,
    extra_imports: str,
    token_dir: Path,
    keystore_path: Path,
) -> None:
    """Pod from Tutorial 10 - tested in the test for tutorial 11."""
    ntbk = jupytext.read(f"{TUTORIALS[10]}.md")
    old_pod_name = SEGMENTATION_POD_NAME
    replacement_pod_name = seg_pod_name
    ntbk = _replace_in_notebook_source_code(old_pod_name, replacement_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "pod = Pod(",
        f'{extra_imports}\npod_keys=_get_pod_keys( Path("{str(token_dir)}")/"pods" / "{replacement_pod_name}")\npod = Pod(pod_keys=pod_keys,',  # noqa: B950,
        ntbk,
    )

    ntbk = _replace_in_notebook_source_code(
        "data_config=segmentation_data_config,",
        f'username="{authentication_user}",\ndata_config=segmentation_data_config,',
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nimport nest_asyncio""",  # noqa: B950,
        ntbk,
    )
    with open(f"{TUTORIALS[10]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()
    importlib.import_module(f"ipynb.fs.full.{TUTORIALS[10]}")


@tutorial_test
@tutorial_11_test
def test_tutorial_11(
    authentication_password: str,
    authentication_user: str,
    census_income_pod_name: str,
    extra_imports: str,
    keystore_path: Path,
    seg_pod_name: str,
    token_dir: Path,
) -> None:
    """Test for tutorial 11- training_a_custom_segmentation_model.md."""
    # set up pod
    pod_process = multiprocessing.Process(
        target=start_segmentation_pod,
        name="Segmentation_Pod_Runner",
        args=(
            authentication_password,
            seg_pod_name,
            authentication_user,
            extra_imports,
            token_dir,
            keystore_path,
        ),
    )
    pod_process.start()
    time.sleep(30)  # give the pods time to spin up
    # assert pod_process.is_alive()

    # change the markdown notebooks to ipynb and replace relevant fields
    ntbk = jupytext.read(f"{TUTORIALS[11]}.md")
    ntbk = _replace_in_notebook_source_code(SEGMENTATION_POD_NAME, seg_pod_name, ntbk)
    ntbk = _replace_in_notebook_source_code(
        "import nest_asyncio",
        f"""from pathlib import Path\nimport bitfount.hub.helper\nbitfount.hub.helper.BITFOUNT_KEY_STORE = Path("{str(keystore_path)}")\nfrom MyCustomSegmentationModel import MyCustomSegmentationModel\nimport nest_asyncio""",  # noqa: B950
        ntbk,
    )
    ntbk = _replace_in_notebook_source_code(
        "schema = get_pod_schema(pod_identifier)",
        "from bitfount import BitfountSchema\nschema = BitfountSchema."
        f"load_from_file('{str(TEST_SCHEMAS_PATH)}/data-schema-segmentation-data-demo.yaml')\n"  # noqa: B950
        f"schema.tables[0].name = '{seg_pod_name}'",
        ntbk,
    )
    for cell in ntbk.cells:
        if (
            cell.cell_type == "code"
            and "class MyCustomSegmentationModel" in cell.source
        ):
            with open("MyCustomSegmentationModel.py", mode="w", encoding="utf-8") as f:
                imports = """import torch\nfrom torch import nn as nn\nfrom torch.nn import functional as F\nfrom bitfount import SoftDiceLoss, SEGMENTATION_METRICS, PyTorchBitfountModel, PyTorchClassifierMixIn\n"""  # noqa: B950
                model = imports + cell.source

                f.write(model)
            cell.source = ""
    with open(f"{TUTORIALS[11]}.ipynb", mode="w", encoding="utf-8") as f:
        nbformat.write(ntbk, f)

    # Patch OIDC challenge
    session_patcher, oidc_patcher = _patch_authentication_code(
        authentication_user, authentication_password, str(token_dir)
    )
    session_patcher.start()
    oidc_patcher.start()

    # Execute the notebook
    notebook = importlib.import_module(  # noqa: B950, F841 # this will be used when updating the tests
        f"ipynb.fs.full.{TUTORIALS[11]}"
    )
    # terminate processes
    session_patcher.stop()
    oidc_patcher.stop()
    pod_process.terminate()



"""Useful functions for concurrency in tests.

In particular, running a modeller and multiple local workers.
"""
import asyncio
from asyncio import Task
from concurrent.futures import ThreadPoolExecutor
import logging
import threading
from typing import Any, Coroutine, List, Literal, Optional, TypeVar, Union

from bitfount import _Modeller
from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from bitfount.federated.worker import _Worker
from bitfount.utils.concurrency_utils import await_threading_event

_logger = logging.getLogger(__name__)

_R = TypeVar("_R")


def _modeller_run(
    modeller: _Modeller, pod_identifiers: List[str], stop: threading.Event
) -> Union[Literal[False], Optional[Any]]:
    """Run _Modeller.run_async() synchronously.

    Designed to be the target of a thread.
    """
    result = asyncio.run(_async_modeller_run(modeller, pod_identifiers, stop))
    _logger.info("Finished modeller_run")
    return result


async def _async_modeller_run(
    modeller: _Modeller, pod_identifiers: List[str], stop: threading.Event
) -> Union[Literal[False], Optional[Any]]:
    """Run _Modeller.run_async() and monitor for stop event."""
    result = await _run_coro_with_stop(
        coro=modeller.run_async(pod_identifiers, True),
        coro_task_name="modeller_run_async_task",
        stop_event=stop,
        stop_event_task_name="modeller_run_stop",
    )

    _logger.info(f"modeller_run_async_task done, result: {result}")
    return result


def _worker_run(worker: _Worker, stop: threading.Event) -> None:
    """Run _Worker.run() synchronously.

    Designed to be the target of a thread.
    """
    asyncio.run(_async_worker_run(worker, stop))
    _logger.info(f"Finished worker_run for worker {worker.pod_identifier}")


async def _async_worker_run(worker: _Worker, stop: threading.Event) -> None:
    """Run _Worker.run() and monitor for stop event.

    Also handles the mailbox listening.
    """
    await _run_coro_with_stop(
        coro=_run_func_and_listen_to_mailbox(worker.run(), worker.mailbox),
        coro_task_name=f"worker_{worker.pod_identifier}_run_async_task",
        stop_event=stop,
        stop_event_task_name=f"worker_{worker.pod_identifier}_run_stop",
    )
    _logger.info(f"worker_{worker.pod_identifier}_run_async_task done")


async def _run_coro_with_stop(
    coro: Coroutine[Any, Any, _R],
    coro_task_name: str,
    stop_event: threading.Event,
    stop_event_task_name: str,
) -> _R:
    """Run a coroutine whilst monitoring for a thread-safe stop event."""
    # Create explicit tasks so we can target them later.
    run_task = asyncio.create_task(
        coro,
        name=coro_task_name,
    )
    stop_task = asyncio.create_task(
        await_threading_event(stop_event, stop_event_task_name),
        name=stop_event_task_name,
    )

    try:
        # Run until either run_task is complete or stop event is set.
        aws: List[Task] = [run_task, stop_task]
        await asyncio.wait(aws, return_when=asyncio.FIRST_COMPLETED)
    except BaseException as e:
        _logger.exception(e)
        raise e
    finally:
        # Find out which task is complete and cancel the other one as needed
        if not run_task.done():
            _logger.error(f"{run_task.get_name()} did not finish, cancelling")
            run_task.cancel()
            try:
                await run_task
            except asyncio.CancelledError:
                pass

        if not stop_task.done():
            _logger.warning(
                f"{stop_task.get_name()} did not finish, cancelling but not awaiting"
            )
            stop_task.cancel()

    return run_task.result()


async def run_local_modeller_and_workers(
    modeller: _Modeller,
    workers: List[_Worker],
) -> Union[Literal[False], Optional[Any]]:
    """Run a pre-configured modeller and workers to task completion.

    Args:
        modeller: The modeller to run.
        workers: The list of workers to run.

    Returns:
        The results of the modeller execution.
    """
    stop_event = threading.Event()
    with ThreadPoolExecutor(max_workers=(1 + len(workers))) as pool:
        try:
            # Create futures for the modeller and workers running.
            main_loop = asyncio.get_running_loop()
            modeller_fut = main_loop.run_in_executor(
                pool,
                _modeller_run,
                modeller,
                [worker.mailbox.pod_identifier for worker in workers],
                stop_event,
            )
            worker_futs = [
                main_loop.run_in_executor(pool, _worker_run, worker, stop_event)
                for worker in workers
            ]

            # Execute until all are done or an exception is raised.
            try:
                modeller_results, *_ = await asyncio.gather(modeller_fut, *worker_futs)
            except BaseException as e:
                _logger.exception(e)

                # Tell other tasks to stop execution within their own threads
                _logger.error("Setting stop_event")
                stop_event.set()

                # Cancel tasks (if already executing won't stop execution,
                # stop_event does that).
                # Can't use gather_task.cancel() as that is already "done" by this point
                to_await: List[asyncio.Future] = []

                # Cancel modeller
                if not modeller_fut.done():
                    _logger.error(
                        f"Cancelling outstanding modeller future {modeller_fut=}"
                    )
                    modeller_fut.cancel()
                    to_await.append(modeller_fut)
                else:
                    _logger.info(f"Modeller task already finished {modeller_fut=}")

                # Cancel workers
                for fut in worker_futs:
                    if not fut.done():
                        _logger.error(f"Cancelling outstanding worker future {fut=}")
                        fut.cancel()
                        to_await.append(fut)
                    else:
                        _logger.info(f"Worker future already finished {fut=}")

                # Await those cancelled tasks
                for fut in to_await:
                    try:
                        _logger.error(f"Waiting for cancelled future to finish {fut=}")
                        await fut
                    except asyncio.CancelledError:
                        pass

                _logger.error("All tasks cancelled")
                raise e
        finally:
            # Guarantee it is set, even if not exceptions
            if not stop_event.is_set():
                _logger.info("Setting stop_event")
                stop_event.set()

        _logger.debug("End of thread pool execution")

    _logger.debug("Fully out of thread pool execution")

    return modeller_results



"""Handles Modeller sending training requests and receiving pod responses."""
from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from grpc import RpcError

from bitfount import config
from bitfount.federated.encryption import _AESEncryption
from bitfount.federated.exceptions import BitfountTaskStartError
from bitfount.federated.logging import _federate_logger, _get_federated_logger
from bitfount.federated.transport.base_transport import (
    Handler,
    MessageRetrievalError,
    _BaseMailbox,
    _send_aes_encrypted_message,
)
from bitfount.federated.transport.handlers import _AsyncMultipleResponsesHandler
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _DecryptedBitfountMessage,
    _MessageService,
)
from bitfount.federated.transport.types import (
    _OIDCAuthFlowResponse,
    _OIDCClientID,
    _PodDeviceCodeDetails,
)
from bitfount.federated.transport.utils import _average_training_metrics
from bitfount.federated.types import (
    _RESPONSE_MESSAGES,
    SerializedProtocol,
    _PodResponseType,
    _TaskRequestMessageGenerator,
)
from bitfount.types import _JSONDict, _SAMLResponse, _SerializedWeights, _StrAnyDict

logger = _get_federated_logger(__name__)

_DEFAULT_TASK_RESPONSE_TIMEOUT: Final[int] = 5 * 60  # 5 minutes
_DEFAULT_OIDC_CLIENT_IDS_TIMEOUT: Final[int] = 5 * 60  # 5 minutes
_SOFT_LIMIT_MESSAGE_TIMEOUT: Final[int] = config.BITFOUNT_ONLINE_CHECK_SOFT_LIMIT


@dataclass
class _WorkerMailboxDetails:
    """Mailbox details for a specific task/worker on a pod.

    Used by the modeller to encapsulate details of worker mailboxes.

    Attributes:
        pod_identifier: The parent pod's identifier.
        public_key: The parent pod's public key.
        mailbox_id: The mailbox ID for this specific task/worker.
        aes_encryption_key: The encryption key to use for this specific task/worker.
    """

    pod_identifier: str
    public_key: RSAPublicKey
    mailbox_id: str
    aes_encryption_key: bytes


class _ModellerMailbox(_BaseMailbox):
    """Handles message interactions with pods."""

    def __init__(
        self,
        mailbox_id: str,
        worker_mailboxes: Mapping[str, _WorkerMailboxDetails],
        task_id: str,
        message_service: _MessageService,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ):
        """Creates a new ModellerMailbox.

        Note that the preferred way to get a new ModellerMailbox is by calling
        ModellerMailbox.send_task_requests() which will instantiate the correct
        ModellerMailbox for you.

        Args:
            mailbox_id: The mailbox ID for this modeller mailbox.
            worker_mailboxes: A mapping of pod identifiers to worker mailbox details
                              for the pods/workers that will be involved in this task.
            task_id: The ID for the task this mailbox belongs to.
            message_service: The underlying message service.
            handlers: Optional. A set of handlers to initialise with.
        """
        super().__init__(
            mailbox_id=mailbox_id, message_service=message_service, handlers=handlers
        )
        self.worker_mailboxes: Dict[str, _WorkerMailboxDetails] = dict(worker_mailboxes)
        self._pod_identifiers: Set[str] = set(worker_mailboxes.keys())
        self._task_id = task_id

        self.accepted_worker_mailboxes: Dict[str, _WorkerMailboxDetails] = {}

        self._setup_federated_logging()
        self._setup_online_status_handler()

    @property
    def task_id(self) -> str:
        """The task ID of the task associated with this mailbox."""
        return self._task_id

    ############################
    # Task Setup Phase Methods #
    ############################
    @classmethod
    async def send_task_requests(
        cls,
        serialized_protocol: SerializedProtocol,
        pod_public_keys: Mapping[str, RSAPublicKey],
        task_request_msg_gen: _TaskRequestMessageGenerator,
        message_service: _MessageService,
    ) -> _ModellerMailbox:
        """Sends task requests, such as training requests, to pods.

        Appropriate mailboxes will be created for the modeller and worker mailboxes
        for the pods.

        Args:
            serialized_protocol: The serialized protocol to use for the task.
            pod_public_keys: A mapping of pod identifiers to their public keys for all
                the pods involved in this task request.
            task_request_msg_gen: A callable which will generate a task request message
                appropriate to the chosen verification method.
            message_service: The underlying message service to use.

        Returns:
            The created modeller mailbox for this task.
        """
        modeller_mailbox_id, worker_mailboxes, task_id = await cls._send_task_requests(
            serialized_protocol,
            pod_public_keys,
            task_request_msg_gen=task_request_msg_gen,
            message_service=message_service,
        )
        modeller_mailbox = cls(
            mailbox_id=modeller_mailbox_id,
            worker_mailboxes=worker_mailboxes,
            task_id=task_id,
            message_service=message_service,
        )
        return modeller_mailbox

    @classmethod
    async def _send_task_requests(
        cls,
        serialized_protocol: SerializedProtocol,
        pod_public_keys: Mapping[str, RSAPublicKey],
        task_request_msg_gen: _TaskRequestMessageGenerator,
        message_service: _MessageService,
    ) -> Tuple[str, Dict[str, _WorkerMailboxDetails], str]:
        """Manage sending of task requests, such as training requests, to pods.

        Args:
            serialized_protocol: The serialized protocol to use for the task.
            pod_public_keys: A mapping of pod identifiers to their public keys for all
                the pods involved in this task request.
            task_request_msg_gen: A callable which will generate a task request message
                appropriate to the chosen verification method.
            message_service: The underlying message service to use.

        Returns:
            Tuple of:
                - (str) modeller mailbox ID
                - (dict) of pod identifier to worker mailbox details
                - (str) the task ID
        """
        # Shorthand the pod identifiers for ease of use
        pod_identifiers: List[str] = list(pod_public_keys)

        # Generate encryption keys and then task request messages for each pod.
        aes_key_per_pod = {
            pod_identifier: _AESEncryption.generate_key()
            for pod_identifier in pod_identifiers
        }
        task_request_per_pod: Dict[str, bytes] = {
            pod_identifier: task_request_msg_gen(
                serialized_protocol,
                pod_identifiers,
                aes_key_per_pod[pod_identifier],
                pod_public_key,
            )
            for pod_identifier, pod_public_key in pod_public_keys.items()
        }

        # Send this message to each pod and receive the modeller's mailbox ID
        # and the mailbox IDs of all the pods in the task.
        try:
            (
                modeller_mailbox_id,
                worker_mailbox_ids,
                task_id,
            ) = await message_service.setup_communication_with_pods(
                task_request_per_pod
            )
            logger.info(f"Sent task requests to {pod_identifiers}")
        except RpcError as err:
            logger.error(
                f"Failed to start task with pods: {pod_identifiers}. Error: {err}"
            )
            raise BitfountTaskStartError(
                f"Failed to start task with pods: {pod_identifiers}"
            )

        return (
            modeller_mailbox_id,
            {
                pod_identifier: _WorkerMailboxDetails(
                    pod_identifier,
                    pod_public_keys[pod_identifier],
                    worker_mailbox_id,
                    aes_key_per_pod[pod_identifier],
                )
                for pod_identifier, worker_mailbox_id in worker_mailbox_ids.items()
            },
            task_id,
        )

    async def get_oidc_client_ids(
        self, timeout: Optional[int] = _DEFAULT_OIDC_CLIENT_IDS_TIMEOUT
    ) -> Dict[str, _OIDCClientID]:
        """Receive OIDC client ID responses from pods.

        These will be the first step in OIDC-related auth flows, showing the modeller
        that the pods have received their request and are ready to start the flow.

        Returns:
            A mapping of pod identifier to their client ID as a dataclass.
        """
        oidc_client_ids: Dict[str, _OIDCClientID] = {}

        def _oidc_client_id_message_handler(message: _BitfountMessage) -> None:
            logger.info(f"Received OIDC Client ID from {message.sender}")
            decrypted_msg = self._decrypt_message(message)
            oidc_client_ids[decrypted_msg.sender] = _OIDCClientID.deserialize(
                decrypted_msg.body
            )

        # Create handler for processing the group of expected responses. Using it
        # as a context manager guarantees that the handlers are correctly applied
        # and removed.
        with _AsyncMultipleResponsesHandler(
            handler=_oidc_client_id_message_handler,
            message_types=_BitfountMessageType.OIDC_CHALLENGE,
            mailbox=self,
            responders=self._pod_identifiers,
        ) as multi_response:
            try:
                # Wait for all responses to have been received or until the
                # timeout expires.
                await multi_response.wait_for_responses(timeout=timeout)
            except MessageRetrievalError as err:
                logger.error(
                    f"Error receiving responses from all pods to the OIDC phase "
                    f"of the task request: {err}"
                )
                raise BitfountTaskStartError("Failed to start task with all pods.")

        if not oidc_client_ids:
            logger.error("No OIDC client id retreived from message handler")
            raise ValueError("No OIDC client id retreived from message handler")

        return oidc_client_ids

    async def send_oidc_auth_flow_responses(
        self,
        oidc_responses: Dict[str, _OIDCAuthFlowResponse],
    ) -> None:
        """Send response for OIDC Authorization Code Flow."""
        for pod_id, response in oidc_responses.items():
            pod_mailbox = self.worker_mailboxes[pod_id]
            await _send_aes_encrypted_message(
                response.serialize(),
                pod_mailbox.aes_encryption_key,
                self.message_service,
                message_type=_BitfountMessageType.OIDC_AFC_PKCE_RESPONSE,
                recipient=pod_mailbox.pod_identifier,
                recipient_mailbox_id=pod_mailbox.mailbox_id,
                sender=self.message_service.username,
                sender_mailbox_id=self.mailbox_id,
                task_id=self._task_id,
            )

    async def send_oidc_device_code_responses(
        self, device_code_details: Dict[str, _PodDeviceCodeDetails]
    ) -> None:
        """Send response for OIDC Device Code Flow."""
        for pod_id, details in device_code_details.items():
            pod_mailbox = self.worker_mailboxes[pod_id]
            await _send_aes_encrypted_message(
                details.serialize(),
                pod_mailbox.aes_encryption_key,
                self.message_service,
                message_type=_BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE,
                recipient=pod_mailbox.pod_identifier,
                recipient_mailbox_id=pod_mailbox.mailbox_id,
                sender=self.message_service.username,
                sender_mailbox_id=self.mailbox_id,
                task_id=self._task_id,
            )

    async def get_saml_challenges(
        self, timeout: int = _DEFAULT_TASK_RESPONSE_TIMEOUT
    ) -> List[_DecryptedBitfountMessage]:
        """Process incoming SAML Challenges.

        Incoming responses are awaited as a group until all are received or timeout
        is reached. Responses are expected from all pods assigned at mailbox init.

        Returns:
            A list of received SAML challenges. Note that there is no notion of
            "pod order" in this list, the elements are in the order they are received.
        """
        saml_challenges = []

        def _saml_challenge_message_handler(message: _BitfountMessage) -> None:
            """Simple handler that saves messages to closured dict."""
            logger.info(
                f"Received message with type: {message.message_type} "
                f"from sender: {message.sender}"
            )
            saml_challenges.append(self._decrypt_message(message))

        # Create handler for processing the group of expected responses. Using it
        # as a context manager guarantees that the handlers are correctly applied
        # and removed.
        with _AsyncMultipleResponsesHandler(
            handler=_saml_challenge_message_handler,
            message_types=_BitfountMessageType.SAML_REQUEST,
            mailbox=self,
            responders=self._pod_identifiers,
        ) as multi_response:
            try:
                # Wait for all responses to have been received or until the
                # timeout expires.
                await multi_response.wait_for_responses(timeout=timeout)
            except MessageRetrievalError as err:
                logger.error(
                    f"Error receiving responses from all pods to the task request: "
                    f"{err}"
                )
                raise BitfountTaskStartError("Failed to start task with all pods.")

        return saml_challenges

    async def process_task_request_responses(
        self, timeout: int = _DEFAULT_TASK_RESPONSE_TIMEOUT
    ) -> Dict[str, _WorkerMailboxDetails]:
        """Process incoming responses to a task request.

        Incoming responses are awaited as a group until all are received or timeout
        is reached. Responses are expected from all pods assigned at mailbox init.

        Returns:
            The subdict of this mailbox's worker mailboxes that corresponds to
            pods that accepted the task.
        """
        # Set the responses all to None to begin with and replace them with
        # messages as we receive them.
        response_messages: Dict[str, Optional[_DecryptedBitfountMessage]] = {
            pod_identifier: None for pod_identifier in self._pod_identifiers
        }

        def _response_message_handler(message: _BitfountMessage) -> None:
            """Simple handler that saves messages to closured dict."""
            logger.info(
                f"Received message with type: {message.message_type} "
                f"from sender: {message.sender}"
            )
            response_messages[message.sender] = self._decrypt_message(message)

        # Create handler for processing the group of expected responses. Using it
        # as a context manager guarantees that the handlers are correctly applied
        # and removed.
        with _AsyncMultipleResponsesHandler(
            handler=_response_message_handler,
            message_types=[
                _BitfountMessageType.JOB_ACCEPT,
                _BitfountMessageType.JOB_REJECT,
            ],
            mailbox=self,
            responders=self._pod_identifiers,
        ) as multi_response:
            try:
                # Wait for all responses to have been received or until the
                # timeout expires.
                await multi_response.wait_for_responses(timeout=timeout)
            except MessageRetrievalError as err:
                logger.error(
                    f"Error receiving responses from all pods to the task request: "
                    f"{err}"
                )
                raise BitfountTaskStartError("Failed to start task with all pods.")

        accepted_mailbox_details = await self._handle_task_responses(response_messages)

        # Set on the attribute and return as well
        self.accepted_worker_mailboxes = accepted_mailbox_details
        return accepted_mailbox_details

    async def _handle_task_responses(
        self,
        response_messages: MutableMapping[str, Optional[_DecryptedBitfountMessage]],
    ) -> Dict[str, _WorkerMailboxDetails]:
        # Want to track the various responses separately, but only care about
        # the details of the accepted ones.
        accepted_task_worker_mailboxes: Dict[str, _WorkerMailboxDetails] = {}
        # These are effectively only used for counting,
        # so could be replaced with counters.
        # The only reason for creating a list of messages here
        # was in case we wanted to log these errors in future
        rejected_tasks: List[_DecryptedBitfountMessage] = []
        ignored_tasks: List[None] = []

        # Check through each of the responses
        for pod_identifier, response in response_messages.items():
            if response:
                pod_identifier = response.sender

                # Handle ACCEPT response
                if _PodResponseType.ACCEPT.name in response.body:
                    logger.info(f"Pod '{pod_identifier}' accepted request")

                    mailbox_details = self.worker_mailboxes[pod_identifier]
                    accepted_task_worker_mailboxes[pod_identifier] = mailbox_details

                    logger.debug(
                        f"Pod {pod_identifier} " f"mailbox id is: {mailbox_details}"
                    )

                # Handle REJECT response (regardless of what form that reject takes)
                else:
                    rejected_tasks.append(response)

                    # Process different forms of rejection
                    for response_type in response.body:
                        logger.error(
                            f"Received rejection from {pod_identifier}. "
                            f"{_RESPONSE_MESSAGES[_PodResponseType[response_type]]}"  # noqa: B950
                        )

            # Handle cases where response didn't arrive
            else:
                response = cast(None, response)
                ignored_tasks.append(response)

        logger.info(
            f"{len(accepted_task_worker_mailboxes)} task(s) accepted, "
            f"{len(rejected_tasks)} rejection(s), "
            f"{len(ignored_tasks)} pod(s) did not respond in time."
        )

        return accepted_task_worker_mailboxes

    async def send_saml_responses(
        self, saml_response: _SAMLResponse, pod_mailbox: _WorkerMailboxDetails
    ) -> None:
        """Send a SAML response to a pod."""
        await _send_aes_encrypted_message(
            saml_response,
            pod_mailbox.aes_encryption_key,
            self.message_service,
            message_type=_BitfountMessageType.SAML_RESPONSE,
            recipient=pod_mailbox.pod_identifier,
            recipient_mailbox_id=pod_mailbox.mailbox_id,
            sender=self.message_service.username,
            sender_mailbox_id=self.mailbox_id,
            task_id=self._task_id,
        )

    ################################
    # End Task Setup Phase Methods #
    ################################

    ##############################
    # Task Running Phase Methods #
    ##############################
    async def _send_to_all_pods_aes_encrypt(
        self, object_to_send: Any, message_type: _BitfountMessageType
    ) -> None:
        """Send message to all pods involved in a training task, AES encrypted.

        Args:
            object_to_send: Body of the message (not encrypted)
            message_type: The type of message to send
        """
        for mailbox in self.accepted_worker_mailboxes.values():
            await _send_aes_encrypted_message(
                object_to_send,
                mailbox.aes_encryption_key,
                self.message_service,
                message_type=message_type,
                recipient=mailbox.pod_identifier,
                recipient_mailbox_id=mailbox.mailbox_id,
                sender_mailbox_id=self.mailbox_id,
                sender=self.message_service.username,
                task_id=self._task_id,
            )

    async def send_training_iteration_complete_update(
        self, training_complete: bool
    ) -> None:
        """Sends whether training is complete or not to the workers."""
        logger.debug(f"Sending TRAINING_COMPLETE from {self.mailbox_id}")
        await self._send_to_all_pods_aes_encrypt(
            training_complete, _BitfountMessageType.TRAINING_COMPLETE
        )

    async def send_task_start_message(self) -> None:
        """Sends task start message to the workers.

        Note: The message is not important here, the message type is.
        """
        await self._send_to_all_pods_aes_encrypt(None, _BitfountMessageType.TASK_START)

    async def send_task_complete_message(self) -> None:
        """Sends task complete message to the workers.

        Note: The message is not important here, the message type is.
        """
        logger.debug(f"Sending TASK_COMPLETE message to workers from {self.mailbox_id}")
        await self._send_to_all_pods_aes_encrypt(
            None, _BitfountMessageType.TASK_COMPLETE
        )

    def _decrypt_message(self, message: _BitfountMessage) -> _DecryptedBitfountMessage:
        """Decrypt received message using this mailbox's AES keys.

        Args:
            message: Received message to decrypt.

        Returns:
            The decrypted message body.
        """
        return message.decrypt(self.worker_mailboxes[message.sender].aes_encryption_key)

    async def get_evaluation_results_from_workers(
        self, timeout: Optional[int] = None
    ) -> _StrAnyDict:
        """Get evaluation results from workers."""
        logger.info("Waiting to receive results from Pods...")
        all_eval_results: _StrAnyDict = {}

        # Create light-weight handler to append to shared list
        def evaluation_results_handler(message: _BitfountMessage) -> None:
            logger.debug(f"Receiving evaluation results from worker {message.sender}")
            eval_results = self._decrypt_message(message).body
            all_eval_results[message.sender] = eval_results

        # We use `self` rather than `self.modeller_mailbox` as the mailbox below
        # because this is ensures things are correctly delegated.
        with _AsyncMultipleResponsesHandler(
            handler=evaluation_results_handler,
            message_types=_BitfountMessageType.EVALUATION_RESULTS,
            mailbox=self,
            responders=self.accepted_worker_mailboxes.keys(),
        ) as response_handler:
            await response_handler.wait_for_responses(timeout=timeout)

        return all_eval_results

    async def log(self, message: Mapping[str, object]) -> None:
        """Log message to all pods involved in task."""
        await self._send_to_all_pods_aes_encrypt(
            message, _BitfountMessageType.LOG_MESSAGE
        )

    def _setup_federated_logging(self) -> None:
        """Set up federated logging."""
        _federate_logger(self)

        def log_message_handler(message: _BitfountMessage) -> None:
            """Locally logs the log message that has been received from a pod."""
            log_message_wrapper: _DecryptedBitfountMessage = self._decrypt_message(
                message
            )
            log_message: _JSONDict = log_message_wrapper.body
            # We prepend the name of the pod to the log message
            log_message["msg"] = f"{message.sender} Pod: {log_message['msg']}"
            # We remove the `federated` key to avoid recursively sending a federated
            # log message on both the Modeller and Worker sides
            log_message.pop("federated")
            logger.handle(logging.makeLogRecord(log_message))

        self.register_handler(_BitfountMessageType.LOG_MESSAGE, log_message_handler)

    def _setup_online_status_handler(self) -> None:
        """Respond to online status requests from Pods."""

        async def status_request_handler(message: _BitfountMessage) -> None:
            """Responds to an ONLINE_CHECK request with an ONLINE_RESPONSE."""
            logger.info(f"Informing {message.sender} that we are still online.")

            # We use the message service sending directly as we don't want to
            # re-encrypt the already encrypted body, we just want to send it back
            # to the worker.
            await self.message_service.send_message(
                _BitfountMessage(
                    message_type=_BitfountMessageType.ONLINE_RESPONSE,
                    body=message.body,
                    recipient=message.sender,
                    recipient_mailbox_id=message.sender_mailbox_id,
                    sender=self.message_service.username,
                    sender_mailbox_id=self.mailbox_id,
                    task_id=self._task_id,
                ),
                already_packed=True,
            )

        self.register_handler(
            _BitfountMessageType.ONLINE_CHECK,
            status_request_handler,
            high_priority=True,
        )

    ##################################
    # End Task Running Phase Methods #
    ##################################


async def _send_model_parameters(
    model_parameters: _SerializedWeights, modeller_mailbox: _ModellerMailbox
) -> None:
    """Sends model parameters to the workers."""
    # Deliberate access to private method here as that method shouldn't be used in
    # any other context than transport layer access.
    # noinspection PyProtectedMember
    await modeller_mailbox._send_to_all_pods_aes_encrypt(
        model_parameters, _BitfountMessageType.MODEL_PARAMETERS
    )


async def _send_psi_dataset_modeller(
    dataset: List[str], modeller_mailbox: _ModellerMailbox
) -> None:
    """Sends psi datasets to the worker."""
    await modeller_mailbox._send_to_all_pods_aes_encrypt(
        dataset, _BitfountMessageType.PSI_DATASET
    )


def _psi_dataset_handler(
    modeller_mailbox: _ModellerMailbox,
    psi_datasets_str: List[Tuple[List[str], List[str]]],
) -> Callable[[_BitfountMessage], None]:
    """Handle receiving of PSI datasets from workers.

    Appends them to the `psi_datasets_str` list.

    Note that there is no notion of "worker order" in this list, the elements are
    in the order they are received.
    """
    # Create light-weight handler to append to shared list
    def psi_dataset_handler(message: _BitfountMessage) -> None:
        logger.debug(f"Receiving psi dataset from worker {message.sender}")
        # Deliberate access to private method here as that method shouldn't be used in
        # any other context than transport layer access.
        # noinspection PyProtectedMember
        psi_results = modeller_mailbox._decrypt_message(message).body
        psi_datasets_str.append(psi_results)

    return psi_dataset_handler


async def _get_psi_datasets_from_workers(
    modeller_mailbox: _ModellerMailbox, timeout: Optional[int] = None
) -> List[Tuple[List[str], List[str]]]:
    """Get psi datasets from workers.

    Note that there is no notion of "worker order" in this list, the elements are
    in the order they are received.
    """
    psi_datasets_str: List[Tuple[List[str], List[str]]] = []
    psi_dataset_handler = _psi_dataset_handler(modeller_mailbox, psi_datasets_str)
    # We use `self` rather than `self.modeller_mailbox` as the mailbox below
    # because this is ensures things are correctly delegated.
    with _AsyncMultipleResponsesHandler(
        handler=psi_dataset_handler,
        message_types=_BitfountMessageType.PSI_DATASET,
        mailbox=modeller_mailbox,
        responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
    ) as response_handler:
        await response_handler.wait_for_responses(timeout=timeout)
    return psi_datasets_str


def _public_key_handler(
    modeller_mailbox: _ModellerMailbox, public_key: List[bytes]
) -> Callable[[_BitfountMessage], None]:
    """Public key handler.

    Will mutate the passed in sequence by appending responses to it.
    Should only be used for the PrivateSetIntersection protocol.

    Note that there is no notion of "worker order" in this list, the elements are
    in the order they are received.
    """

    def public_key_handler(message: _BitfountMessage) -> None:
        logger.debug(f"Receiving the public key from worker {message.sender}")
        # Deliberate access to private method here as that method shouldn't be used in
        # any other context than transport layer access.
        # noinspection PyProtectedMember
        public_key_ = modeller_mailbox._decrypt_message(message).body
        public_key.append(public_key_)

    return public_key_handler


async def _get_public_key(
    modeller_mailbox: _ModellerMailbox, timeout: Optional[int] = None
) -> List[bytes]:
    """Get public key from worker.

    Used for the PrivateSetIntersection protocol.

    Note that there is no notion of "pod order" in this list, the elements are in
    the order they are received.
    """
    public_key_lst: List[bytes] = []
    public_key_handler = _public_key_handler(modeller_mailbox, public_key_lst)
    with _AsyncMultipleResponsesHandler(
        handler=public_key_handler,
        message_types=_BitfountMessageType.KEY_EXCHANGE,
        mailbox=modeller_mailbox,
        responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
    ) as response_handler:
        non_responders = await response_handler.wait_for_responses(timeout=timeout)
        if non_responders:
            logger.info(
                f"The following did not send training metrics in time: {non_responders}"
            )
    return public_key_lst


def _training_metrics_handler(
    modeller_mailbox: _ModellerMailbox,
    training_metrics: MutableSequence[Mapping[str, str]],
) -> Callable[[_BitfountMessage], None]:
    """Training metrics handler.

    Will mutate the passed in sequence by appending responses to it.

    Note that there is no notion of "worker order" in this list, the elements are
    in the order they are received.
    """

    def training_metrics_handler(message: _BitfountMessage) -> None:
        logger.debug(f"Receiving training metrics update from worker {message.sender}")
        # Deliberate access to private method here as that method shouldn't be used in
        # any other context than transport layer access.
        # noinspection PyProtectedMember
        single_training_metrics: Mapping[str, str] = modeller_mailbox._decrypt_message(
            message
        ).body
        training_metrics.append(single_training_metrics)

    return training_metrics_handler


async def _get_training_metrics_from_workers(
    modeller_mailbox: _ModellerMailbox,
    timeout: Optional[int] = None,
) -> Dict[str, float]:
    """Get average training metrics from workers."""
    training_metrics: List[Mapping[str, str]] = []
    training_metrics_handler = _training_metrics_handler(
        modeller_mailbox, training_metrics
    )
    with _AsyncMultipleResponsesHandler(
        handler=training_metrics_handler,
        message_types=_BitfountMessageType.TRAINING_METRICS,
        mailbox=modeller_mailbox,
        responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
    ) as response_handler:
        non_responders = await response_handler.wait_for_responses(timeout=timeout)
        if non_responders:
            logger.info(
                f"The following did not send training metrics in time: {non_responders}"
            )

    # Find the average metrics for those who responded and return
    averaged_training_metrics = _average_training_metrics(training_metrics)
    return averaged_training_metrics


def _parameter_updates_handler(
    modeller_mailbox: _ModellerMailbox,
    weight_updates: MutableMapping[str, _SerializedWeights],
) -> Callable[[_BitfountMessage], None]:
    """Parameter update handler.

    Will mutate the passed in mapping by appending responses to it.
    """

    def parameter_update_handler(message: _BitfountMessage) -> None:
        logger.debug(f"Receiving parameter update from worker {message.sender}")
        # Deliberate access to private method here as that method shouldn't be used in
        # any other context than transport layer access.
        # noinspection PyProtectedMember
        weight_update = modeller_mailbox._decrypt_message(message).body
        sender = message.sender
        weight_updates[sender] = weight_update

    return parameter_update_handler


async def _get_parameter_updates_from_workers(
    modeller_mailbox: _ModellerMailbox, timeout: Optional[int] = None
) -> Dict[str, _SerializedWeights]:
    """Get model parameter updates from workers.

    Args:
        modeller_mailbox: The modeller mailbox.
        timeout: The timeout for the request.

    Returns:
        A dictionary of the form {worker_name: weight_update}.
    """
    weight_updates: Dict[str, _SerializedWeights] = {}
    parameter_updates_handler = _parameter_updates_handler(
        modeller_mailbox, weight_updates
    )

    with _AsyncMultipleResponsesHandler(
        handler=parameter_updates_handler,
        message_types=_BitfountMessageType.TRAINING_UPDATE,
        mailbox=modeller_mailbox,
        responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
    ) as response_handler:
        await response_handler.wait_for_responses(timeout=timeout)

    return weight_updates



"""Custom exceptions for the runners package."""
from bitfount import BitfountError


class PlugInAlgorithmError(BitfountError):
    """Raised if the specified algorithm in not found in the plugins."""

    pass



"""Manages the federated communication and training of models.

Federated algorithm plugins can also be imported from this package.
"""
import importlib as _importlib
import inspect as _inspect
import pkgutil as _pkgutil
from typing import List

from bitfount.config import (
    BITFOUNT_FEDERATED_PLUGIN_PATH as _BITFOUNT_FEDERATED_PLUGIN_PATH,
)
from bitfount.federated import algorithms
from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.algorithms.base import BaseAlgorithmFactory
from bitfount.federated.algorithms.column_avg import ColumnAverage
from bitfount.federated.algorithms.compute_intersection_rsa import (
    ComputeIntersectionRSA,
)
from bitfount.federated.algorithms.model_algorithms.evaluate import ModelEvaluation
from bitfount.federated.algorithms.model_algorithms.federated_training import (
    FederatedModelTraining,
)
from bitfount.federated.algorithms.model_algorithms.inference import ModelInference
from bitfount.federated.algorithms.model_algorithms.train_and_evaluate import (
    ModelTrainingAndEvaluation,
)
from bitfount.federated.algorithms.private_sql_query import PrivateSqlQuery
from bitfount.federated.algorithms.sql_query import SqlQuery
from bitfount.federated.authorisation_checkers import IdentityVerificationMethod
from bitfount.federated.early_stopping import FederatedEarlyStopping
from bitfount.federated.exceptions import (
    AggregatorError,
    BitfountTaskStartError,
    DecryptError,
    EncryptError,
    EncryptionError,
    MessageHandlerNotFoundError,
    MessageRetrievalError,
    PodConnectFailedError,
    PodNameError,
    PodRegistrationError,
    PrivateSqlError,
    SecureShareError,
)
from bitfount.federated.helper import combine_pod_schemas
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.modeller import _Modeller
from bitfount.federated.pod import Pod
from bitfount.federated.pod_keys_setup import PodKeys
from bitfount.federated.privacy.differential import DPModellerConfig, DPPodConfig
from bitfount.federated.protocols.model_protocols.federated_averaging import (
    FederatedAveraging,
)
from bitfount.federated.protocols.psi import PrivateSetIntersection
from bitfount.federated.protocols.results_only import ResultsOnly
from bitfount.federated.roles import Role
from bitfount.federated.secure import SecureShare
from bitfount.federated.shim import BackendTensorShim
from bitfount.federated.transport import MAXIMUM_GRPC_MESSAGE_SIZE_BYTES
from bitfount.federated.transport.config import (
    PRODUCTION_MESSAGE_SERVICE_URL,
    MessageServiceConfig,
)
from bitfount.federated.types import AggregatorType, AlgorithmType, ProtocolType
from bitfount.utils import _get_module_from_file

_logger = _get_federated_logger(__name__)

__all__: List[str] = [
    "Aggregator",
    "AggregatorError",
    "AggregatorType",
    "AlgorithmType",
    "BackendTensorShim",
    "BitfountModelReference",
    "BitfountTaskStartError",
    "ColumnAverage",
    "DPModellerConfig",
    "DPPodConfig",
    "DecryptError",
    "EncryptError",
    "EncryptionError",
    "FederatedAveraging",
    "FederatedEarlyStopping",
    "FederatedModelTraining",
    "IdentityVerificationMethod",
    "MAXIMUM_GRPC_MESSAGE_SIZE_BYTES",
    "MessageHandlerNotFoundError",
    "MessageRetrievalError",
    "MessageServiceConfig",
    "ModelEvaluation",
    "ModelInference",
    "ModelTrainingAndEvaluation",
    "_Modeller",
    "Pod",
    "PodConnectFailedError",
    "PodKeys",
    "PodNameError",
    "PodRegistrationError",
    "PrivateSetIntersection",
    "PrivateSqlError",
    "PrivateSqlQuery",
    "ProtocolType",
    "PRODUCTION_MESSAGE_SERVICE_URL",
    "ResultsOnly",
    "Role",
    "ComputeIntersectionRSA",
    "SecureAggregator",
    "SecureShare",
    "SecureShareError",
    "SqlQuery",
    "combine_pod_schemas",
]

# Import all concrete implementations of BaseAlgorithmFactory in the algorithms
# subdirectory as well as algorithms plugins

for _module_info in _pkgutil.walk_packages(
    path=algorithms.__path__ + [str(_BITFOUNT_FEDERATED_PLUGIN_PATH / "algorithms")],
):
    try:
        _module = _importlib.import_module(f"{algorithms.__name__}.{_module_info.name}")
    # Also catches `ModuleNotFoundError` which subclasses `ImportError`
    # Try to import the module from the plugin directory if it's not found in the
    # algorithms directory
    except ImportError:
        try:
            _module = _get_module_from_file(
                _BITFOUNT_FEDERATED_PLUGIN_PATH
                / "algorithms"
                / f"{_module_info.name}.py",
            )
        except ImportError:
            _logger.debug(f"Error importing module {_module_info.name}")
    finally:
        for _, cls in _inspect.getmembers(_module, _inspect.isclass):
            if issubclass(cls, BaseAlgorithmFactory) and not _inspect.isabstract(cls):
                globals().update({cls.__name__: getattr(_module, cls.__name__)})
                __all__.append(cls.__name__)
            # There are too many false positives if we don't restrict classes to those
            # that inherit from BaseAlgorithmFactory for it to be a useful log message
            elif (
                issubclass(cls, BaseAlgorithmFactory)
                and cls.__name__ != "BaseAlgorithmFactory"
            ):
                _logger.debug(
                    f"Found class {cls.__name__} in module {_module_info.name} which "
                    f"did not fully implement BaseAlgorithmFactory. Skipping."
                )


# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Utilities for logging and warning messages functionality."""
from contextlib import contextmanager
from datetime import datetime
import logging
from pathlib import Path
import re
import sys
from typing import Final, Generator, List, Optional, Union
import warnings

from bitfount.config import BITFOUNT_LOG_TO_FILE, BITFOUNT_LOGS_DIR

# Timestamp when this module is read in. Guarantees that any file-loggers will refer
# to the same file.
_log_file_time: Final = datetime.now().strftime("%Y-%m-%d-%H%M%S")


@contextmanager
def filter_stderr(to_filter: Union[str, re.Pattern]) -> Generator[None, None, None]:
    """Filter stderr messages emitted within this context manager.

    Will remove any messages where the start matches the filter pattern whilst allowing
    any other messages to go through.

    Args:
        to_filter: Regex pattern to match the start of messages to be filtered.
    """
    # Compile regex pattern if not already done
    reg_to_filter: re.Pattern[str]
    if isinstance(to_filter, str):
        reg_to_filter = re.compile(to_filter)
    else:
        reg_to_filter = to_filter

    # Store previous stderr.write() method
    _stderr_write = sys.stderr.write

    def _write(s: str) -> int:
        """Override write() method of stderr."""
        if reg_to_filter.match(s):
            # Do nothing, write 0 bytes
            return 0
        else:
            return _stderr_write(s)

    # mypy_reason: mypy is overzealous with functions being assigned to instance
    #              methods as it cannot easily determine the type of the callable
    #              between bound and unbound. "type: ignore" is the recommended
    #              workaround.
    #              See: https://github.com/python/mypy/issues/2427
    try:
        sys.stderr.write = _write  # type: ignore[assignment] # Reason: see comment
        yield
    finally:
        # Reset stderr.write() method
        sys.stderr.write = _stderr_write  # type: ignore[assignment] # Reason: see comment # noqa: B950


def _get_bitfount_logdir(subdir_name: Optional[str] = None) -> Path:
    """Get the directory that Bitfount logs should be written to.

    This will be BITFOUNT_LOGS_DIR unless subdir_name is specified in which case
    it will be that subdirectory within BITFOUNT_LOGS_DIR.
    """
    # Establish log directory and create it if it doesn't exist
    parent_logfile_dir = BITFOUNT_LOGS_DIR

    if subdir_name:
        logfile_dir = parent_logfile_dir / subdir_name
    else:
        logfile_dir = parent_logfile_dir

    logfile_dir.mkdir(parents=True, exist_ok=True)
    return logfile_dir


def _get_bitfount_log_file_handler(
    log_file_subdir_name: Optional[str] = None,
) -> logging.FileHandler:
    """Get a FileHandler pre-configured for Bitfount style.

    Will create the log file in BITFOUNT_LOGS_DIR unless log_file_subdir_name is
    specified in which case it will be created in that subdirectory within
    BITFOUNT_LOGS_DIR.

    Log-level will be DEBUG.
    """
    logfile_dir = _get_bitfount_logdir(log_file_subdir_name)

    # Set file logging configuration
    file_handler = logging.FileHandler(f"{logfile_dir}/{_log_file_time}.log")
    file_log_formatter = logging.Formatter(
        "%(asctime)s:"
        " %(threadName)s:"
        " [%(levelname)-8s]"
        " %(name)s:%(filename)s:%(lineno)d:"
        " %(message)s"
    )
    file_handler.setFormatter(file_log_formatter)
    file_handler.setLevel(logging.DEBUG)

    return file_handler


def _configure_logging() -> None:
    """Configure logging and third-party loggers to adhere to Bitfount style."""
    # Set up logging to capture any `warnings` module issues raised.
    logging.captureWarnings(True)

    # Handle pytorch_lightning loggers if pytorch_lightning is enabled
    try:
        # Need to import here to guarantee that the pytorch_lightning logger is
        # set up before we override it.
        import pytorch_lightning  # noqa: F401

        # Change level to WARNING and above, ensure only sent to file, not to
        # stream, ensure not propagated beyond this logger.
        pytorch_logger = logging.getLogger("pytorch_lightning")
        pytorch_logger.setLevel(logging.WARNING)
        pytorch_logger.handlers = []
        pytorch_logger.addHandler(_get_bitfount_log_file_handler())
        pytorch_logger.propagate = False

        # Filter out warnings related to IterableDataset (as not an issue for us
        # as we guarantee only one worker); should only be processed once, not
        # per-batch.
        warnings.filterwarnings(
            action="once",
            message=r"Your `IterableDataset` has `__len__` defined.",
            category=UserWarning,
            module="pytorch_lightning.utilities",
        )

        # Filter out warnings from pytorch_lightning, redirecting them to file only.
        # Other warnings are printed to stderr AND logged to file.
        # Avoid propagating to the root logger.
        warnings_logger = logging.getLogger("py.warnings")
        warnings_logger.propagate = False

        non_lightning_warning_handler = logging.StreamHandler(sys.stderr)
        non_lightning_warning_handler.addFilter(is_not_pytorch_lightning_warning)
        warnings_logger.addHandler(non_lightning_warning_handler)

        all_warnings_handler = _get_bitfount_log_file_handler()
        warnings_logger.addHandler(all_warnings_handler)
    except ImportError:
        pass


def setup_loggers(
    loggers: List[logging.Logger], log_file_dir_name: Optional[str] = None
) -> List[logging.Logger]:
    """Set up loggers with console and file handlers.

    Creates a logfile in 'logs' directory with the current date and time and outputs all
    logs at the "DEBUG" level. Also outputs logs to stdout at the "INFO" level. A common
    scenario is to attach handlers only to the root logger, and to let propagation take
    care of the rest.

    Args:
        loggers: The logger(s) to set up
        log_file_dir_name: Creates a subdirectory inside BITFOUNT_LOGS_DIR
            if provided. Defaults to None.

    Returns:
        A list of updated logger(s).
    """
    handlers: List[logging.Handler] = []

    # If logging to file is enabled, create appropriate FileHandler
    if BITFOUNT_LOG_TO_FILE:
        file_handler = _get_bitfount_log_file_handler(log_file_dir_name)
        handlers.append(file_handler)

    # Set console logging configuration
    console_handler = logging.StreamHandler(sys.stdout)
    console_log_formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s]: %(message)s", datefmt="%H:%M:%S"
    )
    console_handler.setFormatter(console_log_formatter)
    console_handler.setLevel(logging.INFO)
    handlers.append(console_handler)

    # Cannot use `logger` as iter-variable as shadows outer name.
    for i_logger in loggers:
        # Clear any existing handler/filter configuration
        i_logger.handlers = []
        i_logger.filters = []

        # Set base level to DEBUG and ensure messages are not duplicated
        i_logger.setLevel(logging.DEBUG)
        i_logger.propagate = False

        # Add handlers to loggers
        for handler in handlers:
            i_logger.addHandler(handler)

    return loggers


# PyTorch-Lightning-related utilities
def is_not_pytorch_lightning_warning(record: logging.LogRecord) -> int:
    """Returns 0 if is a warning generated by PyTorch Lightning, 1 otherwise."""
    msg: str = record.getMessage()

    # warnings generates a log message of form "<file_path>:<line_no> ..." so should
    # be able to only filter on lightning mentions in file path.
    file_path: str = msg.split(":", maxsplit=1)[0]
    if "pytorch_lightning" in file_path:
        return 0  # i.e. should not be logged
    else:
        return 1  # i.e. should be logged



"""Algorithms for remote processing of data.

Federated algorithm plugins can also be imported from this package.
"""
import pkgutil as _pkgutil

from bitfount.config import (
    BITFOUNT_FEDERATED_PLUGIN_PATH as _BITFOUNT_FEDERATED_PLUGIN_PATH,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.utils import _get_module_from_file

_logger = _get_federated_logger(__name__)


# Create `algorithms` plugin subdir if it doesn't exist
_algorithms_plugin_path = _BITFOUNT_FEDERATED_PLUGIN_PATH / "algorithms"
_algorithms_plugin_path.mkdir(parents=True, exist_ok=True)


def _log_import_error(pkg: str) -> None:
    _logger.error(f"Issue importing {pkg}")


# Add algorithms plugin modules to the `algorithms` namespace alongside the existing
# built-in algorithms modules. This is not essential, but it allows users to import
# the entire plugin module as opposed to just the built-in class which is what is done
# in the `bitfount.federated` __init__ module.
for _module_info in _pkgutil.walk_packages(
    [str(_algorithms_plugin_path)],
):
    try:
        _module = _get_module_from_file(
            _algorithms_plugin_path / f"{_module_info.name}.py"
        )
        globals().update({_module.__name__: _module})
        _logger.info(f"Loaded algorithms plugin: {_module_info.name}")
    except ImportError:
        pass



"""This package contains the core Bitfount codebase.

With the exception of `backends` which contains optional extra packages, every
subpackage or standalone module in this package must have an `__init__.py` file that
defines `__all__`.
"""
#   ____  _ _    __                   _
#  | __ )(_) |_ / _| ___  _   _ _ __ | |_
#  |  _ \| | __| |_ / _ \| | | | '_ \| __|
#  | |_) | | |_|  _| (_) | |_| | | | | |_
#  |____/|_|\__|_|  \___/ \__,_|_| |_|\__|

import importlib as _importlib
import logging as _logging
import pkgutil as _pkgutil
import sys as _sys
from types import TracebackType as _TracebackType
from typing import Any as _Any
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Type as _Type
from typing import cast as _cast

from bitfount import (
    config,
    data,
    exceptions,
    federated,
    hub,
    metrics,
    models,
    runners,
    storage,
    transformations,
    types,
    utils,
)
from bitfount.__version__ import __version__  # noqa: F401
from bitfount.config import *  # noqa: F403
from bitfount.data import *  # noqa: F403
from bitfount.exceptions import *  # noqa: F403
from bitfount.federated import *  # noqa: F403
from bitfount.hub import *  # noqa: F403
from bitfount.metrics import *  # noqa: F403
from bitfount.models import *  # noqa: F403
from bitfount.runners import *  # noqa: F403
from bitfount.storage import *  # noqa: F403
from bitfount.transformations import *  # noqa: F403
from bitfount.types import *  # noqa: F403
from bitfount.utils import *  # noqa: F403

__all__: _List[str] = []

from bitfount.utils.logging_utils import _configure_logging

_configure_logging()
_logger = _logging.getLogger(__name__)

# Attempt to import backends if any exist
try:
    import bitfount.backends

    _backends_imported = True
except ModuleNotFoundError:
    _backends_imported = False

# If backends has been successfully imported, attempt to import each individual backend
# and add its __all__ to __all__
if _backends_imported:
    # Find all top-level subpackages in the backends package
    for _module_info in _pkgutil.iter_modules(
        bitfount.backends.__path__, f"{bitfount.backends.__name__}."
    ):
        _module = None
        # Attempt to import backend subpackage
        try:
            _module = _importlib.import_module(_module_info.name)
        except ImportError:
            pass

        # Add backend subpackage's __all__ to __all__
        if _module is not None:
            _imports: _List[str] = []
            try:
                _imports = _module.__dict__["__all__"]
                __all__.extend(_imports)
            except KeyError:
                _logger.error(f"Couldn't import {_module}: __all__ not defined.")

            # Add backend imports defined in __all__ to globals dictionary
            globals().update({k: getattr(_module, k) for k in _imports})


__all__.extend(config.__all__)
__all__.extend(data.__all__)
__all__.extend(exceptions.__all__)
__all__.extend(federated.__all__)
__all__.extend(hub.__all__)
__all__.extend(metrics.__all__)
__all__.extend(models.__all__)
__all__.extend(runners.__all__)
__all__.extend(storage.__all__)
__all__.extend(transformations.__all__)
__all__.extend(types.__all__)
__all__.extend(utils.__all__)

# Currently, due to pdoc's reliance on `__all__`, we must iterate over `__all__`` to
# ignore every import in the documentation otherwise they become duplicated
# https://github.com/pdoc3/pdoc/issues/340
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False

# Set stacktrace dependent on config
if not config.BITFOUNT_DEV_MODE:
    _sys.tracebacklimit = config.BITFOUNT_TB_LIMIT

    # Jupyter/iPython has its own traceback system, so we need to handle that
    # differently
    try:
        # This function will be in scope if we're in jupyter/iPython
        ipython = get_ipython()  # type: ignore[name-defined] # Reason: see comment # noqa: B950, F405

        orig_showtraceback = ipython.showtraceback

        def _limited_showtraceback(
            exc_tuple: _Optional[
                _Tuple[_Type[BaseException], BaseException, _TracebackType]
            ] = None,
            filename: _Any = None,
            tb_offset: _Any = None,
            exception_only: bool = False,
            running_compiled_code: bool = False,
        ) -> _Any:
            """Override ipython.showtraceback so the traceback stack is limited."""
            try:
                try:
                    tb_limit = _sys.tracebacklimit
                except AttributeError:
                    # If tracebacklimit not set, use original implementation
                    return orig_showtraceback(
                        exc_tuple,
                        filename,
                        tb_offset,
                        exception_only,
                        running_compiled_code,
                    )

                # Otherwise, limit the traceback(s)
                if exc_tuple is None:
                    exc_tuple = _cast(
                        _Tuple[_Type[BaseException], BaseException, _TracebackType],
                        _sys.exc_info(),
                    )
                exc_type, exc_value, tb = exc_tuple

                # Use limit to get Nth entry from back in the traceback
                tb_stack = []
                tb_curr: _Optional[_TracebackType] = tb
                while tb_curr:
                    tb_stack.append(tb_curr)
                    tb_curr = tb_curr.tb_next
                if tb_limit < len(tb_stack):
                    # +1 as IPython interprets the exception itself as the "last frame"
                    # in the stack trace, and we want to at least see where it's raised.
                    nth_tb = tb_stack[-(tb_limit + 1)]
                else:
                    nth_tb = tb

                if nth_tb is not tb:
                    frames_str = "frames" if tb_limit > 1 else "frame"
                    _logger.warning(
                        f"Exception traceback limited to {tb_limit} {frames_str};"
                        f" see debug logs for more details"
                        f" or set BITFOUNT_TB_LIMIT environment variable to a higher"
                        f" number of frames."
                    )

                # Use this in the original implementation
                return orig_showtraceback(
                    (exc_type, exc_value, nth_tb),
                    filename,
                    tb_offset,
                    exception_only,
                    running_compiled_code,
                )
            except Exception:
                print(
                    "Error whilst printing traceback, falling back to default printer",
                    file=_sys.stderr,
                )
                return orig_showtraceback(
                    exc_tuple,
                    filename,
                    tb_offset,
                    exception_only,
                    running_compiled_code,
                )

        ipython.showtraceback = _limited_showtraceback
    except NameError:
        # We are not in a Jupyter/iPython environment
        pass
